from flask_babel import lazy_gettext as _
import requests

import re
import sys

from hiddifypanel.cache import cache


to_gig_d = 1000 * 1000 * 1000


def error(str):
    print(str, file=sys.stderr)


@cache.cache(ttl=60000)
def get_latest_release_url(repo):
    latest_url = requests.get(f'{repo}/releases/latest').url.strip()
    version = latest_url.split('tag/')[1].strip()
    return (latest_url, version)


@cache.cache(ttl=600)
def get_latest_release_version(repo_name):
    try:
        url = f"https://github.com/hiddify/{repo_name}/releases/latest"
        response = requests.head(url, allow_redirects=False)

        location_header = response.headers.get("Location")
        if location_header:
            version = re.search(r"/([^/]+)/?$", location_header)
            if version:
                ver = version.group(1).replace('v', '')
                if ver == "latest":
                    return get_latest_release_version(repo_name.replace("-", ""))
                return ver
    except Exception as e:
        return f'{e}'

    return None



from flask_classful import FlaskView, route
from hiddifypanel import hutils
from hiddifypanel.auth import login_required, current_account, login_user, logout_user, login_by_uuid
from flask import redirect, request, g, render_template, flash, jsonify
from hiddifypanel.hutils.flask import hurl_for
from flask import current_app as app
from flask_babel import lazy_gettext as _
from apiflask import abort
import hiddifypanel.panel.hiddify as hiddify
from hiddifypanel.models import *

from flask_wtf import FlaskForm
import wtforms as wtf
import re


class LoginForm(FlaskForm):
    secret_textbox = wtf.fields.StringField(_(f'login.secret.label'), [wtf.validators.Regexp(
        "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", re.IGNORECASE, _('config.invalid uuid'))], default='',
        description=_(f'login.secret.description'), render_kw={
        'required': "",
        'pattern': "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$",
        'message': _('config.invalid uuid')
    })
    submit = wtf.fields.SubmitField(_('Submit'))


class LoginView(FlaskView):

    # @route("/")
    def index(self, force=None, next=None):
        force_arg = request.args.get('force')
        redirect_arg = request.args.get('redirect')
        username_arg = request.args.get('user') or ''
        if not current_account:

            return render_template('login.html', form=LoginForm())

            # abort(401, "Unauthorized1")

        if redirect_arg:
            return redirect(redirect_arg)
        if hutils.flask.is_admin_proxy_path() and g.account.role in {Role.super_admin, Role.admin, Role.agent}:
            return redirect(hurl_for('admin.Dashboard:index'))
        # if g.user_agent['is_browser'] and hutils.flask.is_client_proxy_path():
        #     return redirect(hurl_for('client.UserView:index'))

        from hiddifypanel.panel.user import UserView
        return UserView().auto_sub()

    def post(self):
        form = LoginForm()
        if form.validate_on_submit():
            uuid = form.secret_textbox.data.strip()
            if login_by_uuid(uuid, hutils.flask.is_admin_proxy_path()):
                return redirect(f'/{g.proxy_path}/')
        hutils.flask.flash(_('config.validation-error'), 'danger')  # type: ignore
        return render_template('login.html', form=LoginForm())

    @ route("/l/<path:path>/")
    @ route("/l/<path:path>")
    @ route("/l/")
    @ route("/l")
    def basic(self, path=None):
        if path:
            redirect_arg = f"/{g.proxy_path}/{path}"
        else:
            redirect_arg = request.args.get('next')

        if not current_account or (not request.headers.get('Authorization')):
            username = request.authorization.username if request.authorization else ''

            loginurl = hurl_for('common_bp.LoginView:index', next=redirect_arg, user=username)
            if g.user_agent['is_browser'] and request.headers.get('Authorization') or (current_account and len(username) > 0 and current_account.username != username):
                hutils.flask.flash(_('Incorrect Password'), 'error')  # type: ignore
                logout_user()
                g.__account_store = None
                # hutils.flask.flash(request.authorization.username, 'error')
                return redirect(loginurl)

            return render_template("redirect.html", url=loginurl), 401
            # abort(401, "Unauthorized1")
        if redirect_arg:
            return redirect(redirect_arg)

        if hutils.flask.is_admin_proxy_path() and g.account.role in {Role.super_admin, Role.admin, Role.agent}:
            return redirect(hurl_for('admin.Dashboard:index'))

        if g.user_agent['is_browser'] and hutils.flask.is_client_proxy_path():
            return redirect(hurl_for('client.UserView:index'))

        from hiddifypanel.panel.user import UserView
        # return redirect(url_for("user.")) UserView().auto_sub()

    # @route('/<uuid:uuid>/<path:path>')
    # @route('/<uuid:uuid>/')

    # def uuid(self, uuid, path=''):
    #     proxy_path = hiddify.flask.get_proxy_path_from_url(request.url)
    #     g.__account_store = None
    #     uuid = str(uuid)
    #     if proxy_path == hconfig(ConfigEnum.proxy_path_client):
    #         g.__account_store = User.by_uuid(uuid)
    #         path = f'client/{path}'
    #     elif proxy_path == hconfig(ConfigEnum.proxy_path_admin):
    #         g.__account_store = AdminUser.by_uuid(uuid)
    #     if not g.account:
    #         abort(403)
    #     if not g.user_agent['is_browser'] and proxy_path == hconfig(ConfigEnum.proxy_path_client):
    #         userview = UserView()
    #         if "all.txt" in path:
    #             return userview.all_configs()
    #         if 'singbox.json' in path:
    #             return userview.singbox()
    #         if 'full-singbox.json' in path:
    #             return userview.full_singbox()
    #         if 'clash' in path:
    #             splt = path.split("/")
    #             meta_or_normal = 'meta' if splt[-2] == 'meta' else 'normal'
    #             typ = splt[-1].split('.yml')[0]
    #             return userview.clash_config(meta_or_normal=meta_or_normal, typ=typ)
    #         return userview.force_sub()

    #     login_user(g.account, force=True)

    #     return redirect(f"/{proxy_path}/{path}")

    @ route('/<secret_uuid>/manifest.webmanifest')
    def create_pwa_manifest(self):
        domain = request.host
        name = (domain if hutils.flask.is_admin_panel_call() else g.account.name)
        return jsonify({
            "name": f"Hiddify {name}",
            "short_name": f"{name}"[:12],
            "theme_color": "#f2f4fb",
            "background_color": "#1a1b21",
            "display": "standalone",
            "scope": f"/",
            "start_url": hiddify.get_account_panel_link(g.account, domain) + "?pwa=true",
            "description": "Hiddify, for a free Internet",
            "orientation": "any",
            "icons": [
                {
                    "src": hutils.flask.static_url_for(filename='images/hiddify-dark.png'),
                    "sizes": "512x512",
                    "type": "image/png",
                    "purpose": "maskable any"
                }
            ]
        })



from apiflask import APIBlueprint

bp = APIBlueprint("api_panel", __name__, url_prefix="/<proxy_path>/api/v2/panel/", enable_openapi=True)


def init_app(app):

    with app.app_context():
        from .info import PanelInfoApi
        from .ping_pong import PingPongApi
        bp.add_url_rule('/info/', view_func=PanelInfoApi)
        bp.add_url_rule('/ping/', view_func=PingPongApi)
    app.register_blueprint(bp)



from loguru import logger
import socket

from hiddifypanel.models import AdminUser, User, hconfig, ConfigEnum, ChildMode, set_hconfig, Domain, Proxy, StrConfig, BoolConfig, Child, ChildMode
from hiddifypanel import hutils
from hiddifypanel.panel import hiddify
from hiddifypanel.panel import usage
from hiddifypanel.database import db
from hiddifypanel.cache import cache

# import schmeas
from hiddifypanel.panel.commercial.restapi.v2.parent.schema import *
from hiddifypanel.panel.commercial.restapi.v2.child.schema import *

from .api_client import NodeApiClient, NodeApiErrorSchema
# region private


def __get_register_data_for_api(name: str, mode: ChildMode) -> RegisterInputSchema:

    register_data = RegisterInputSchema()
    register_data.unique_id = hconfig(ConfigEnum.unique_id)
    register_data.name = name  # type: ignore
    register_data.mode = mode  # type: ignore

    panel_data = RegisterDataSchema()  # type: ignore
    panel_data.admin_users = [admin_user.to_schema() for admin_user in AdminUser.query.all()]  # type: ignore
    panel_data.users = [user.to_schema() for user in User.query.all()]  # type: ignore
    panel_data.domains = [domain.to_schema() for domain in Domain.query.all()]  # type: ignore
    panel_data.proxies = [proxy.to_schema() for proxy in Proxy.query.all()]  # type: ignore
    panel_data.hconfigs = [*[u.to_schema() for u in StrConfig.query.all()], *[u.to_schema() for u in BoolConfig.query.all()]]  # type: ignore
    register_data.panel_data = panel_data

    return register_data


def __get_sync_data_for_api() -> SyncInputSchema:
    sync_data = SyncInputSchema()
    sync_data.domains = [domain.to_schema() for domain in Domain.query.all()]  # type: ignore
    sync_data.proxies = [proxy.to_schema() for proxy in Proxy.query.all()]  # type: ignore
    sync_data.hconfigs = [*[u.to_schema() for u in StrConfig.query.all()], *[u.to_schema() for u in BoolConfig.query.all()]]  # type: ignore

    return sync_data


def __get_parent_panel_url() -> str:
    url = 'https://' + f"{hconfig(ConfigEnum.parent_domain).removesuffix('/')}/{hconfig(ConfigEnum.parent_admin_proxy_path).removesuffix('/')}"
    return url

# endregion


def is_registered() -> bool:
    '''Checks if the current parent registered as a child'''
    try:
        logger.debug("Checking if current panel is registered with parent")
        base_url = __get_parent_panel_url()
        if not base_url:
            return False
        payload = ChildStatusInputSchema()
        payload.child_unique_id = hconfig(ConfigEnum.unique_id)

        res = NodeApiClient(base_url).post('/api/v2/parent/status/', payload, ChildStatusOutputSchema)
        if isinstance(res, NodeApiErrorSchema):
            logger.error(f"Error while checking if current panel is registered with parent: {res.msg}")
            return False

        if res['existance']:
            return True
        return False
    except Exception as e:
        logger.error(f"Error while checking if current panel is registered with parent")
        logger.exception(e)
        return False


def register_to_parent(name: str, apikey: str, mode: ChildMode = ChildMode.remote) -> bool:
    # get parent link its format is "https://panel.hiddify.com/<admin_proxy_path>/"
    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Parent url is empty")
        return False

    payload = __get_register_data_for_api(name, mode)
    res = NodeApiClient(p_url, apikey).put('/api/v2/parent/register/', payload, RegisterOutputSchema)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while registering to parent: {res.msg}")
        return False

    # TODO: change the bulk_register and such methods to accept models instead of dict
    AdminUser.bulk_register(res['admin_users'], commit=False)
    User.bulk_register(res['users'], commit=False)

    # add new child as parent
    db.session.add(  # type: ignore
        Child(unique_id=res['parent_unique_id'], name=socket.gethostname() or res['parent_unique_id'], mode=ChildMode.parent)
    )

    db.session.commit()  # type: ignore

    logger.success("Successfully registered to parent")
    cache.invalidate_all_cached_functions()
    return True


def sync_with_parent() -> bool:
    # sync usage first
    if not sync_users_usage_with_parent():
        logger.error("Error while syncing with parent: Failed to sync users usage")
        return False

    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Error while syncing with parent: Parent url is empty")
        return False
    payload = __get_sync_data_for_api()
    res = NodeApiClient(p_url).put('/api/v2/parent/sync/', payload, SyncOutputSchema)
    if isinstance(res, NodeApiErrorSchema):
        logger.error(f"Error while syncing with parent: {res.msg}")
        return False
    AdminUser.bulk_register(res['admin_users'], commit=False, remove=True)
    User.bulk_register(res['users'], commit=False, remove=True)
    db.session.commit()  # type: ignore
    logger.success("Successfully synced with parent")
    cache.invalidate_all_cached_functions()
    return True


def sync_users_usage_with_parent() -> bool:
    p_url = __get_parent_panel_url()
    if not p_url:
        logger.error("Parent url is empty")
        return False

    payload = hutils.node.get_users_usage_data_for_api()
    if payload:
        res = NodeApiClient(p_url).put('/api/v2/parent/usage/', payload, UsageInputOutputSchema)  # type: ignore
        if isinstance(res, NodeApiErrorSchema):
            logger.error(f"Error while syncing users usage with parent: {res.msg}")
            return False

        # parse usages data
        res = hutils.node.convert_usage_api_response_to_dict(res)  # type: ignore
        usage.add_users_usage_uuid(res, hiddify.get_child(None), True)
        logger.success(f"Successfully synced users usage with parent: {res}")

    return True



from hiddifypanel.panel import hiddify
from flask.views import MethodView

from flask import current_app as app
from flask import g, request
from apiflask import Schema, abort
from apiflask.fields import String, URL, Enum, List, Nested
from flask_babel import lazy_gettext as _
from urllib.parse import urlparse
import user_agents
from strenum import StrEnum
from enum import auto
from hiddifypanel.panel.user.user import get_common_data
from hiddifypanel.hutils.utils import get_latest_release_url
from hiddifypanel import hutils
from hiddifypanel.models.role import Role
from hiddifypanel.auth import login_required
from hiddifypanel.hutils.flask import static_url_for
# region App Api DTOs


class AppInstallType(StrEnum):
    google_play = auto()
    app_store = auto()
    appimage = auto()
    snapcraft = auto()
    microsoft_store = auto()
    apk = auto()
    dmg = auto()
    setup = auto()
    portable = auto()
    other = auto()


class AppInstall(Schema):
    title = String()
    type = Enum(AppInstallType, required=True, description='The platform that provides the app to download')
    url = URL(required=True, description='The url to download the app')


class AppSchema(Schema):
    title = String(required=True, description='The title of the app')
    description = String(required=True, description='The description of the app')
    icon_url = URL(required=True, description='The icon url of the app')
    guide_url = URL(description='The guide url of the app')
    deeplink = URL(required=True, description='The deeplink of the app to imoprt configs')
    install = List(Nested(AppInstall()), required=True, description='The install url of the app')

# this class is not a Data Transfer Object, It's just an enum


class Platform(StrEnum):
    all = auto()
    android = auto()
    ios = auto()
    windows = auto()
    linux = auto()
    mac = auto()
    auto = auto()


class AppInSchema(Schema):
    platform = Enum(
        Platform, load_default=Platform.auto, required=False,
        description='The platform (Operating System) to know what clients should be send. Possible values are: all, android, ios, windows, linux, mac, auto.')
# endregion


class AppAPI(MethodView):
    decorators = [login_required({Role.user})]

    def __init__(self) -> None:
        super().__init__()
        self.hiddify_github_repo = 'https://github.com/hiddify'

        self.user_panel_url = f"https://{urlparse(request.base_url).hostname}/{g.proxy_path}/{g.account.uuid}/"
        self.user_panel_encoded_url = hutils.encode.url_encode(self.user_panel_url)
        c = get_common_data(g.account.uuid, 'new')
        self.subscription_link_url = f"{self.user_panel_url}all.txt?name={c['db_domain'].alias or c['db_domain'].domain}-{c['asn']}&asn={c['asn']}&mode={c['mode']}"
        self.subscription_link_encoded_url = hutils.encode.do_base_64(self.subscription_link_url)
        domain = c['db_domain'].alias or c['db_domain'].domain
        self.profile_title = hutils.encode.url_encode(c['profile_title'])
        # self.clash_all_sites = f"https://{domain}/{g.proxy_path}/clash/all.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_all_{domain}-{c['mode']}"
        # self.clash_foreign_sites = f"https://{domain}/{g.proxy_path}/clash/normal.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_normal_{domain}-{c['mode']}"
        # self.clash_blocked_sites = f"https://{domain}/{g.proxy_path}/clash/lite.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_lite_{c['db_domain'].alias or c['db_domain'].domain}-{c['mode']}"
        # self.clash_meta_all_sites = f"https://{domain}/{g.proxy_path}/clash/meta/all.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_mall_{domain}-{c['mode']}"
        # self.clash_meta_foreign_sites = f"https://{domain}/{g.proxy_path}/clash/meta/normal.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_mnormal_{domain}-{c['mode']}"
        self.clash_meta_blocked_sites = f"https://{domain}/{g.proxy_path}/clash/meta/lite.yml?mode={c['mode']}&asn={c['asn']}&name={c['asn']}_mlite_{domain}-{c['mode']}"

    @app.input(AppInSchema, arg_name='data', location="query")
    @app.output(AppSchema(many=True))
    @login_required({Role.user})
    def get(self, data):
        # parse user agent
        if data['platform'] == Platform.auto:
            platfrom = self.__get_ua_platform()
            if not platfrom:
                abort(400, 'Your selected platform is invalid!')
            self.platform = platfrom
        else:
            self.platform = data['platform']

        # output data
        apps_data = []

        match self.platform:
            case Platform.all:
                apps_data = self.__get_all_apps_dto()
            case Platform.android:
                hiddify_next_dto = self.__get_hiddify_next_app_dto()
                hiddifyng_dto = self.__get_hiddifyng_app_dto()
                v2rayng_dto = self.__get_v2rayng_app_dto()
                hiddify_clash_android_dto = self.__get_hiddify_clash_android_app_dto()
                nekobox_dto = self.__get_nekobox_app_dto()
                apps_data += ([hiddify_next_dto, hiddifyng_dto, v2rayng_dto, hiddify_clash_android_dto, nekobox_dto])
            case Platform.windows:
                hiddify_next_dto = self.__get_hiddify_next_app_dto()
                hiddify_clash_dto = self.__get_hiddify_clash_desktop_app_dto()
                hiddifyn_dto = self.__get_hiddifyn_app_dto()
                apps_data += ([hiddify_next_dto, hiddify_clash_dto, hiddifyn_dto])
            case Platform.ios:
                stash_dto = self.__get_stash_app_dto()
                shadowrocket_dto = self.__get_shadowrocket_app_dto()
                foxray_dto = self.__get_foxray_app_dto()
                streisand_dto = self.__get_streisand_app_dto()
                loon_dto = self.__get_loon_app_dto()
                apps_data += ([streisand_dto, stash_dto, shadowrocket_dto, foxray_dto, loon_dto])
            case Platform.linux:
                hiddify_next_dto = self.__get_hiddify_next_app_dto()
                hiddify_clash_dto = self.__get_hiddify_clash_desktop_app_dto()
                apps_data += ([hiddify_next_dto, hiddify_clash_dto,])
            case Platform.mac:
                hiddify_clash_dto = self.__get_hiddify_clash_desktop_app_dto()
                hiddify_next_dto = self.__get_hiddify_next_app_dto()
                apps_data += ([hiddify_next_dto, hiddify_clash_dto])

        return apps_data

    def __get_ua_platform(self):
        os = user_agents.parse(request.user_agent.string).os.family
        if os == 'Android':
            return Platform.android
        elif os == 'Windows':
            return Platform.windows
        elif os == 'Mac OS X':
            return Platform.mac
        elif os == 'iOS':
            return Platform.ios
        elif 'Linux' in request.user_agent.string and 'X11' or 'Wayland' in request.user_agent.string:
            return Platform.linux

        return None

    def __get_all_apps_dto(self):
        hiddifyn_app_dto = self.__get_hiddifyn_app_dto()
        v2rayng_app_dto = self.__get_v2rayng_app_dto()
        hiddifyng_app_dto = self.__get_hiddifyng_app_dto()
        hiddify_android_app_dto = self.__get_hiddify_clash_android_app_dto()
        foxray_app_dto = self.__get_foxray_app_dto()
        shadowrocket_app_dto = self.__get_shadowrocket_app_dto()
        streisand_app_dto = self.__get_streisand_app_dto()
        loon_app_dto = self.__get_loon_app_dto()
        stash_app_dto = self.__get_stash_app_dto()
        hiddify_clash_app_dto = self.__get_hiddify_clash_desktop_app_dto()
        hiddify_next_app_dto = self.__get_hiddify_next_app_dto()
        return [
            hiddifyn_app_dto, v2rayng_app_dto, hiddifyng_app_dto, hiddify_android_app_dto,
            foxray_app_dto, shadowrocket_app_dto, streisand_app_dto,
            loon_app_dto, stash_app_dto, hiddify_clash_app_dto, hiddify_next_app_dto
        ]

    def __get_app_icon_url(self, app_name):
        base = f'https://{urlparse(request.base_url).hostname}'
        url = ''
        if app_name == _('app.hiddify.next.title'):
            url = base + static_url_for(filename='apps-icon/hiddify_next.ico')
        elif app_name == _('app.hiddifyn.title'):
            url = base + static_url_for(filename='apps-icon/hiddifyn.ico')
        elif app_name == _('app.v2rayng.title'):
            url = base + static_url_for(filename='apps-icon/v2rayng.ico')
        elif app_name == _('app.hiddifyng.title'):
            url = base + static_url_for(filename='apps-icon/hiddifyng.ico')
        elif app_name == _('app.hiddify-clash-android.title'):
            url = base + static_url_for(filename='apps-icon/hiddify_android.ico')
        elif app_name == _('app.foxray.title'):
            url = base + static_url_for(filename='apps-icon/foxray.ico')
        elif app_name == _('app.shadowrocket.title'):
            url = base + static_url_for(filename='apps-icon/shadowrocket.ico')
        elif app_name == _('app.streisand.title'):
            url = base + static_url_for(filename='apps-icon/streisand.ico')
        elif app_name == _('app.loon.title'):
            url = base + static_url_for(filename='apps-icon/loon.ico')
        elif app_name == _('app.stash.title'):
            url = base + static_url_for(filename='apps-icon/stash.ico')
        elif app_name == _('app.hiddify-clash-desktop.title'):
            url = base + static_url_for(filename='apps-icon/hiddify_clash.ico')
        elif app_name == _('app.nekobox.title'):
            url = base + static_url_for(filename='apps-icon/nekobox.ico')

        return url

    def __get_app_install_dto(self, install_type: AppInstallType, url, title=''):
        install_dto = AppInstall()
        install_dto.title = title
        install_dto.type = install_type
        install_dto.url = url
        return install_dto

    def __get_hiddifyn_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.hiddifyn.title')
        dto.description = _('app.hiddifyn.description')
        dto.icon_url = self.__get_app_icon_url(_('app.hiddifyn.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=o9L2sI2T53Q'
        dto.deeplink = f'hiddify://install-sub?url={self.subscription_link_url}'

        ins_url = f'{self.hiddify_github_repo}/HiddifyN/releases/latest/download/HiddifyN.zip'
        dto.install = [self.__get_app_install_dto(AppInstallType.portable, ins_url)]
        return dto

    def __get_nekobox_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.nekobox.title')
        dto.description = _('app.nekobox.description')
        dto.icon_url = self.__get_app_icon_url(_('app.nekobox.title'))
        dto.guide_url = ''
        dto.deeplink = f'clash://install-config?url={self.user_panel_encoded_url}'

        latest_url, version = get_latest_release_url(f'https://github.com/MatsuriDayo/NekoBoxForAndroid')
        ins_url = latest_url.split('releases/')[0] + f'releases/download/{version}/NB4A-{version}-x86_64.apk'
        dto.install = [self.__get_app_install_dto(AppInstallType.apk, ins_url)]
        return dto

    def __get_v2rayng_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.v2rayng.title')
        dto.description = _('app.v2rayng.description')
        dto.icon_url = self.__get_app_icon_url(_('app.v2rayng.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=6HncctDHXVs'
        dto.deeplink = f'v2rayng://install-sub/?url={self.user_panel_encoded_url}'

        # make v2rayng latest version url download
        latest_url, version = get_latest_release_url(f'https://github.com/2dust/v2rayNG/')
        github_ins_url = latest_url.split('releases/')[0] + f'releases/download/{version}/v2rayNG_{version}.apk'
        google_play_ins_url = 'https://play.google.com/store/apps/details?id=com.v2ray.ang'
        dto.install = [self.__get_app_install_dto(AppInstallType.apk, github_ins_url), self.__get_app_install_dto(AppInstallType.google_play, google_play_ins_url)]
        return dto

    def __get_hiddifyng_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.hiddifyng.title')
        dto.description = _('app.hiddifyng.description')
        dto.icon_url = self.__get_app_icon_url(_('app.hiddifyng.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=qDbI72J-INM'
        dto.deeplink = f'hiddify://install-sub/?url={self.user_panel_encoded_url}'

        latest_url, version = get_latest_release_url(f'{self.hiddify_github_repo}/HiddifyNG/')
        github_ins_url = latest_url.split('releases/')[0] + f'releases/download/{version}/HiddifyNG.apk'
        google_play_ins_url = 'https://play.google.com/store/apps/details?id=ang.hiddify.com'
        dto.install = [self.__get_app_install_dto(AppInstallType.apk, github_ins_url), self.__get_app_install_dto(AppInstallType.google_play, google_play_ins_url)]
        return dto

    def __get_hiddify_clash_android_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.hiddify-clash-android.title')
        dto.description = _('app.hiddify-clash-android.description')
        dto.icon_url = self.__get_app_icon_url(_('app.hiddify-clash-android.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=mUTfYd1_UCM'
        dto.deeplink = f'clash://install-config/?url={self.user_panel_encoded_url}'
        latest_url, version = get_latest_release_url(f'{self.hiddify_github_repo}/HiddifyClashAndroid/')
        ins_url = latest_url.split('releases/')[0] + f'releases/download/{version}/hiddify-{version}-meta-alpha-universal-release.apk'
        dto.install = [self.__get_app_install_dto(AppInstallType.apk, ins_url),]
        return dto

    def __get_foxray_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.foxray.title')
        dto.description = _('app.foxray.description')
        dto.icon_url = self.__get_app_icon_url(_('app.foxray.title'))
        dto.guide_url = ''
        dto.deeplink = f'https://yiguo.dev/sub/add/?url={hutils.encode.do_base_64(self.subscription_link_encoded_url)}#{self.profile_title}'

        ins_url = 'https://apps.apple.com/us/app/foxray/id6448898396'
        dto.install = [self.__get_app_install_dto(AppInstallType.app_store, ins_url),]
        return dto

    def __get_shadowrocket_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.shadowrocket.title')
        dto.description = _('app.shadowrocket.description')
        dto.icon_url = self.__get_app_icon_url(_('app.shadowrocket.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=F2bC_mtbYmQ'
        dto.deeplink = f'sub://{hutils.encode.do_base_64(self.user_panel_url)}'

        ins_url = 'https://apps.apple.com/us/app/shadowrocket/id932747118'
        dto.install = [self.__get_app_install_dto(AppInstallType.app_store, ins_url),]
        return dto

    def __get_streisand_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.streisand.title')
        dto.description = _('app.streisand.description')
        dto.icon_url = self.__get_app_icon_url(_('app.streisand.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=jaMkZTLH2QY'
        dto.deeplink = f'streisand://import/{self.user_panel_url}#{self.profile_title}'
        ins_url = 'https://apps.apple.com/app/id6450534064'
        dto.install = [self.__get_app_install_dto(AppInstallType.app_store, ins_url),]
        return dto

    def __get_loon_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.loon.title')
        dto.description = _('app.loon.description')
        dto.icon_url = self.__get_app_icon_url(_('app.loon.title'))
        dto.guide_url = ''
        dto.deeplink = f'loon://import?nodelist={self.user_panel_encoded_url}'
        ins_url = 'https://apps.apple.com/app/id1373567447'
        dto.install = [self.__get_app_install_dto(AppInstallType.app_store, ins_url)]

    def __get_stash_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.stash.title')
        dto.description = _('app.stash.description')
        dto.icon_url = self.__get_app_icon_url(_('app.stash.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=D0Xv54nRSY8'
        dto.deeplink = f'clash://install-config/?url={self.user_panel_encoded_url}'

        ins_url = 'https://apps.apple.com/us/app/stash-rule-based-proxy/id1596063349'
        dto.install = [self.__get_app_install_dto(AppInstallType.app_store, ins_url),]
        return dto

    def __get_hiddify_clash_desktop_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.hiddify-clash-desktop.title')
        dto.description = _('app.hiddify-clash-desktop.description')
        dto.icon_url = self.__get_app_icon_url(_('app.hiddify-clash-desktop.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=omGIz97mbzM'
        dto.deeplink = f'clash://install-config/?url={self.user_panel_encoded_url}'
        dto.install = []

        # make hiddify clash latest version url download
        latest_url, version = get_latest_release_url(f'{self.hiddify_github_repo}/hiddifydesktop')
        version = version.replace('v', '')

        platform = self.platform

        if self.platform == Platform.all:
            platform = [Platform.windows, Platform.linux, Platform.mac]
        
        def get_link(p):
            match p:
                case Platform.windows:
                    ins_url = latest_url.split('releases/')[0] + f'releases/download/v{version}/HiddifyClashDesktop_{version}_x64_en-US.msi'
                    dto.install.append(self.__get_app_install_dto(AppInstallType.setup, ins_url))
                case Platform.linux:
                    ins_url = latest_url.split('releases/')[0] + f'releases/download/v{version}/hiddify-clash-desktop_{version}_amd64.AppImage'
                    dto.install.append(self.__get_app_install_dto(AppInstallType.appimage, ins_url))
                case Platform.mac:
                    ins_url = latest_url.split('releases/')[0] + f'releases/download/v{version}/HiddifyClashDesktop_{version}_x64.dmg'
                    dto.install.append(self.__get_app_install_dto(AppInstallType.dmg, ins_url))

        if isinstance(platform, list):
            for p in platform:
                get_link(p)
        else:
            get_link(platform)

        return dto

    def __get_hiddify_next_app_dto(self):
        dto = AppSchema()
        dto.title = _('app.hiddify.next.title')
        dto.description = _('app.hiddify.next.description')
        dto.icon_url = self.__get_app_icon_url(_('app.hiddify.next.title'))
        dto.guide_url = 'https://www.youtube.com/watch?v=vUaA1AEUy1s'
        dto.deeplink = f'hiddify://import/{self.user_panel_url}'

        # availabe installatoin types
        installation_types = []
        if self.platform == Platform.all:
            installation_types = [AppInstallType.apk, AppInstallType.google_play, AppInstallType.setup, AppInstallType.portable, AppInstallType.appimage, AppInstallType.dmg]
        else:
            match self.platform:
                case Platform.android:
                    installation_types = [AppInstallType.apk, AppInstallType.google_play]
                case Platform.windows:
                    installation_types = [AppInstallType.setup, AppInstallType.portable]
                case Platform.linux:
                    installation_types = [AppInstallType.appimage]
                case Platform.mac:
                    installation_types = [AppInstallType.dmg]

        install_dtos = []
        for install_type in installation_types:
            install_dto = AppInstall()
            ins_url = ''
            match install_type:
                case AppInstallType.apk:
                    ins_url = f'{self.hiddify_github_repo}/hiddify-next/releases/latest/download/Hiddify-Android-universal.apk'
                case AppInstallType.google_play:
                    ins_url = 'https://play.google.com/store/apps/details?id=app.hiddify.com'
                case AppInstallType.setup:
                    ins_url = f'{self.hiddify_github_repo}/hiddify-next/releases/latest/download/Hiddify-Windows-Setup-x64.exe'
                case AppInstallType.portable:
                    ins_url = f'{self.hiddify_github_repo}/hiddify-next/releases/latest/download/Hiddify-Windows-Portable-x64.zip'
                case AppInstallType.appimage:
                    ins_url = f'{self.hiddify_github_repo}/hiddify-next/releases/latest/download/Hiddify-Linux-x64.AppImage'
                case AppInstallType.dmg:
                    ins_url = f'{self.hiddify_github_repo}/hiddify-next/releases/latest/download/Hiddify-MacOS.dmg'

            install_dto = self.__get_app_install_dto(install_type, ins_url)
            install_dtos.append(install_dto)

        dto.install = install_dtos
        return dto



from .login import LoginView
from apiflask import APIBlueprint
bp = APIBlueprint("common_bp", __name__, url_prefix="/<proxy_path>/", template_folder="templates", enable_openapi=False)


def init_app(app):
    LoginView.register(bp, route_base="/")
    app.register_blueprint(bp)
    app.register_blueprint(bp, name=f"child_{bp.name}", url_prefix="/<proxy_path>/<int:child_id>/")



from flask import current_app as app
from flask import g
from flask.views import MethodView
from apiflask import abort
from hiddifypanel.auth import login_required
from hiddifypanel.models import *

from . import has_permission
from .schema import AdminSchema, PatchAdminSchema, SuccessfulSchema


class AdminUserApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin})]

    @app.output(AdminSchema)  # type: ignore
    def get(self, uuid):
        admin = AdminUser.by_uuid(uuid) or abort(404, "admin not found")
        if not has_permission(admin):
            abort(403, "You don't have permission to access this admin")
        return admin.to_dict()  # type: ignore

    @app.input(PatchAdminSchema, arg_name='data')  # type: ignore
    @app.output(SuccessfulSchema)  # type: ignore
    def patch(self, uuid, data):
        admin = AdminUser.by_uuid(uuid) or abort(404, "admin not found")
        if not has_permission(admin):
            abort(403, "You don't have permission to access this admin")

        data['uuid'] = uuid
        if not data.get('added_by_uuid'):
            data['added_by_uuid'] = g.account.uuid

        AdminUser.add_or_update(**data)
        return {'status': 200, 'msg': 'ok'}

    @app.output(SuccessfulSchema)  # type: ignore
    def delete(self, uuid):
        admin = AdminUser.by_uuid(uuid) or abort(404, "admin not found")
        if not has_permission(admin):
            abort(403, "You don't have permission to access this admin")
        admin.remove()  # type: ignore
        return {'status': 200, 'msg': 'ok'}



#!/usr/bin/env python3
from flask import render_template, request, jsonify, redirect, g
from flask_wtf.file import FileField, FileRequired
from flask_bootstrap import SwitchField
from flask_babel import gettext as _
from flask_classful import FlaskView
from urllib.parse import urlparse
from flask_wtf import FlaskForm
from datetime import datetime
import wtforms as wtf
import json


from hiddifypanel.auth import login_required
from hiddifypanel.panel import hiddify
from hiddifypanel.models import *
from hiddifypanel import hutils


class Backup(FlaskView):
    decorators = [login_required({Role.super_admin})]

    def index(self):
        return render_template('backup.html', restore_form=get_restore_form())

    # @route("/backupfile")
    def backupfile(self):
        response = jsonify(
            hiddify.dump_db_to_dict()
        )
        o = urlparse(request.base_url)
        domain = o.hostname
        response.headers.add('Content-disposition', f'attachment; filename=hiddify-{domain}-{datetime.now()}.json')

        return response

    def post(self):

        restore_form = get_restore_form()

        if restore_form.validate_on_submit():
            set_hconfig(ConfigEnum.first_setup, False)
            file = restore_form.restore_file.data
            if isinstance(file, list):
                file = file[0]
            json_data = json.load(file)
            hiddify.set_db_from_json(json_data,
                                     set_users=restore_form.enable_user_restore.data,
                                     set_domains=restore_form.enable_domain_restore.data,
                                     set_settings=restore_form.enable_config_restore.data,
                                     override_unique_id=False,
                                     override_child_unique_id=True,
                                     override_root_admin=restore_form.override_root_admin.data
                                     # replace_owner_admin=restore_form.override_root_admin.data,
                                     )

            # remove default user cause it's not needed (program users wanted)
            if default := User.by_id(1):
                default.remove()

            from flask_babel import refresh
            refresh()
            # return redirect(hutils.flask.hurl_for("admin.Actions:reinstall2"))
            from .Actions import Actions
            g.new_proxy_path = hconfig(ConfigEnum.proxy_path_admin)
            g.force_proxy_path = g.proxy_path
            return Actions().reinstall(complete_install=True, domain_changed=True)
            # # hutils.flask.flash_config_success(full_install=True)
        else:
            hutils.flask.flash(_('Config file is incorrect'), category='error')
        return render_template('backup.html', restore_form=restore_form)


def get_restore_form(empty=False):
    class RestoreForm(FlaskForm):
        restore_file = FileField(_("Restore File"), description=_("Restore File Description"), validators=[FileRequired()])
        enable_config_restore = SwitchField(_("Restore Settings"), description=_("Restore Settings description"), default=False)
        enable_user_restore = SwitchField(_("Restore Users"), description=_("Restore Users description"), default=False)
        enable_domain_restore = SwitchField(_("Restore Domain"), description=_("Restore Domain description"), default=False)
        override_root_admin = SwitchField(_("Override Root Admin"), description=_("It will override the root admin to the current user"), default=False)
        submit = wtf.fields.SubmitField(_('Submit'))

    return RestoreForm(None) if empty else RestoreForm()



def gen_username(model) -> None:
    from hiddifypanel import hutils
    if model.username:
        return
    base_username = model.name or ''
    minimum_username_length = 10

    if len(base_username) < minimum_username_length:
        base_username += hutils.random.get_random_string(minimum_username_length - len(base_username), minimum_username_length)

    if len(base_username) > 100:
        base_username = base_username[0:100]
    model.username = base_username

    while not model.is_username_unique():
        rand_str = hutils.random.get_random_string(2, 4)
        model.username += rand_str


def gen_password(model) -> None:
    from hiddifypanel import hutils
    # TODO: hash the password
    if not model.password or len(model.password) < 16:
        model.password = hutils.random.get_random_password(length=16)


def gen_wg_keys(model) -> None:
    from hiddifypanel import hutils
    if not model.wg_pk or not model.wg_pub or not model.wg_psk:
        model.wg_pk, model.wg_pub, model.wg_psk = hutils.crypto.get_wg_private_public_psk_pair()


def gen_ed25519_keys(model) -> None:
    from hiddifypanel import hutils
    if not model.ed25519_private_key or not model.ed25519_public_key:
        model.ed25519_private_key, model.ed25519_public_key = hutils.crypto.get_ed25519_private_public_pair()



import ipaddress
from hiddifypanel.auth import login_required, current_account

from hiddifypanel.models import *
import re
from flask import g  # type: ignore
from markupsafe import Markup

from flask_babel import gettext as __
from flask_babel import lazy_gettext as _
from hiddifypanel.panel.run_commander import Command, commander
from wtforms.validators import Regexp, ValidationError

from hiddifypanel.models import *
from hiddifypanel.panel import hiddify, cf_api, custom_widgets
from .adminlte import AdminLTEModelView
from hiddifypanel import hutils

from flask import current_app
# Define a custom field type for the related domains


# class ConfigDomainsField(SelectField):
#     def __init__(self, label=None, validators=None,*args, **kwargs):
#         kwargs.pop("allow_blank")
#         super().__init__(label, validators,*args, **kwargs)
#         self.choices=[(d.id,d.domain) for d in Doamin.query.filter(Domain.sub_link_only!=True).all()]


class DomainAdmin(AdminLTEModelView):
    # edit_modal = False
    # create_modal = False
    column_hide_backrefs = False

    list_template = 'model/domain_list.html'
    # edit_modal = True
    form_overrides = {'mode': custom_widgets.EnumSelectField}
    form_widget_args = {
        'description': {
            'rows': 100,
            'style': 'font-family: monospace; direction:ltr'
        }
    }
    column_descriptions = dict(
        domain=_("domain.description"),
        mode=_("Direct mode means you want to use your server directly (for usual use), CDN means that you use your server on behind of a CDN provider."),
        cdn_ip=_("config.cdn_forced_host.description"),
        show_domains=_(
            'You can select the configs with which domains show be shown in the user area. If you select all, automatically, all the new domains will be added for each users.'),
        alias=_('The name shown in the configs for this domain.'),
        servernames=_('config.reality_server_names.description'),
        sub_link_only=_('This can be used for giving your users a permanent non blockable links.'),
        grpc=_('gRPC is a H2 based protocol. Maybe it is faster for you!'))

    # create_modal = True
    can_export = False
    form_widget_args = {'show_domains': {'class': 'form-control ltr'}}
    form_args = {
        'mode': {
            'enum': DomainType},
        'show_domains': {
            'query_factory': lambda: Domain.query.filter(
                Domain.sub_link_only == False),
        },
        'domain': {
            'validators': [
                Regexp(
                    r'^(\*\.)?([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,})$',
                    message=__("Should be a valid domain"))]},
        "cdn_ip": {
            'validators': [
                Regexp(
                    r"(((((25[0-5]|(2[0-4]|1\d|[1-9]|)\d).){3}(25[0-5]|(2[0-4]|1\d|[1-9]|)\d))|^([A-Za-z0-9\-\.]+\.[a-zA-Z]{2,}))[ \t\n,;]*\w{3}[ \t\n,;]*)*",
                    message=__("Invalid IP or domain"))]},
        "servernames": {
            'validators': [
                Regexp(
                    r"^([\w-]+\.)+[\w-]+(,\s*([\w-]+\.)+[\w-]+)*$",
                    re.IGNORECASE,
                    _("Invalid REALITY hostnames"))]}}
    column_list = ["domain", "alias", "mode", "domain_ip", "show_domains"]
    column_editable_list = ["alias"]
    # column_filters=["domain","mode"]
    # form_excluded_columns=['work_with']
    column_searchable_list = ["domain", "mode"]
    column_labels = {
        "domain": _("domain.domain"),
        'sub_link_only': _('Only for sublink?'),
        "mode": _("domain.mode"),
        "cdn_ip": _("config.cdn_forced_host.label"),
        'domain_ip': _('domain.ip'),
        'servernames': _('config.reality_server_names.label'),
        'show_domains': _('Show Domains'),
        'alias': _('Alias'),
        'grpc': _('gRPC')
    }

    form_columns = ['mode', 'domain', 'alias', 'servernames', 'grpc', 'cdn_ip', 'show_domains']

    def _domain_admin_link(view, context, model, name):
        if model.mode == DomainType.fake:
            return Markup(f"<span class='badge'>{model.domain}</span>")
        d = model.domain
        if "*" in d:
            d = d.replace("*", hutils.random.get_random_string(5, 15))
        admin_link = hiddify.get_account_panel_link(g.account, d)
        return Markup(
            f'<div class="btn-group"><a href="{admin_link}" class="btn btn-xs btn-secondary">' + _("admin link") +
            f'</a><a href="{admin_link}" class="btn btn-xs btn-info ltr" target="_blank">{model.domain}</a></div>')

    def _domain_ip(view, context, model, name):
        dip = hutils.network.get_domain_ip(model.domain)
        myip = hutils.network.get_ip(4)
        if myip == dip and model.mode == DomainType.direct:
            badge_type = ''
        elif dip and model.mode != DomainType.direct and myip != dip:
            badge_type = 'warning'
        else:
            badge_type = 'danger'
        res = f'<span class="badge badge-{badge_type}">{dip}</span>'
        if model.sub_link_only:
            res += f'<span class="badge badge-success">{_("SubLink")}</span>'
        return Markup(res)

    def _show_domains_formater(view, context, model, name):
        if not len(model.show_domains):
            return _("All")
        else:
            return Markup(" ".join([hiddify.get_domain_btn_link(d) for d in model.show_domains]))

    column_formatters = {
        'domain_ip': _domain_ip,
        'domain': _domain_admin_link,
        'show_domains': _show_domains_formater
    }

    def search_placeholder(self):
        return f"{_('search')} {_('domain.domain')} {_('domain.mode')}"

    # def on_form_prefill(self, form, id):
        # Get the Domain object being edited
        # domain = self.session.query(Domain).get(id)

        # Pre-select the related domains in the checkbox list
        # form.show_domains = [d.id for d in Domain.query.all()]

    def on_model_change(self, form, model, is_created):
        model.domain = model.domain.lower()

        configs = get_hconfigs()
        for c in configs:
            if "domain" in c and c not in [ConfigEnum.decoy_domain, ConfigEnum.reality_fallback_domain] and c.category != 'hidden':
                if model.domain == configs[c]:
                    raise ValidationError(_("You have used this domain in: ") + _(f"config.{c}.label"))

        for td in Domain.query.filter(Domain.mode == DomainType.reality, Domain.domain != model.domain).all():
            # print(td)
            if td.servernames and (model.domain in td.servernames.split(",")):
                raise ValidationError(_("You have used this domain in: ") + _(f"config.reality_server_names.label") + td.domain)

        if is_created and Domain.query.filter(Domain.domain == model.domain, Domain.child_id == model.child_id).count() > 1:
            raise ValidationError(_("You have used this domain in: "))

        ipv4_list = hutils.network.get_ips(4)
        ipv6_list = hutils.network.get_ips(6)

        if not ipv4_list and not ipv6_list:
            raise ValidationError(_("Couldn't find your ip addresses"))

        if "*" in model.domain and model.mode not in [DomainType.cdn, DomainType.auto_cdn_ip]:
            raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))

        skip_check = "*" in model.domain
        if hconfig(ConfigEnum.cloudflare) and model.mode != DomainType.fake:
            try:
                proxied = model.mode in [DomainType.cdn, DomainType.auto_cdn_ip]
                cf_api.add_or_update_domain(model.domain, str(ipv4_list[0]), "A", proxied=proxied)
                if ipv6_list:
                    cf_api.add_or_update_domain(model.domain, str(ipv6_list[0]), "AAAA", proxied=proxied)

                skip_check = True
            except Exception as e:
                raise ValidationError(__("Can not connect to Cloudflare.") + f' {e}')
        # elif model.mode==DomainType.auto_cdn_ip:
        # if model.alias and not model.alias.replace("_", "").isalnum():
        #     hutils.flask.flash(__("Using alias with special charachters may cause problem in some clients like FairVPN."), 'warning')
            # raise ValidationError(_("You have to add your cloudflare api key to use this feature: "))

        dip = hutils.network.get_domain_ip(model.domain)
        if model.sub_link_only:
            if dip is None:
                raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))  # type: ignore
        elif not skip_check:
            if dip is None:
                raise ValidationError(_("Domain can not be resolved! there is a problem in your domain"))  # type: ignore

            domain_ip_is_same_as_panel = False
            domain_ip_is_same_as_panel |= dip in ipv4_list
            for ipv6 in ipv6_list:
                domain_ip_is_same_as_panel |= ipaddress.ip_address(dip) == ipaddress.ip_address(ipv6)

            if model.mode == DomainType.direct and not domain_ip_is_same_as_panel:
                # hutils.flask.flash(__(f"Domain IP={dip} is not matched with your ip={', '.join(list(map(str, ipv4_list)))} which is required in direct mode"),category='error')
                raise ValidationError(
                    __("Domain IP=%(domain_ip)s is not matched with your ip=%(server_ip)s which is required in direct mode", server_ip=', '.join(list(map(str, ipv4_list))), domain_ip=dip))  # type: ignore

            if domain_ip_is_same_as_panel and model.mode in [DomainType.cdn, DomainType.relay, DomainType.fake, DomainType.auto_cdn_ip]:
                #     # hutils.flask.flash(__(f"In CDN mode, Domain IP={dip} should be different to your ip={', '.join(list(map(str, ipv4_list)))}"), 'warning')
                raise ValidationError(__("In CDN mode, Domain IP=%(domain_ip)s should be different to your ip=%(server_ip)s",
                                      server_ip=', '.join(list(map(str, ipv4_list))), domain_ip=dip))  # type: ignore

            # if model.mode in [DomainType.ss_faketls, DomainType.telegram_faketls]:
            #     if len(Domain.query.filter(Domain.mode==model.mode and Domain.id!=model.id).all())>0:
            #         ValidationError(f"another {model.mode} is exist")

        model.domain = model.domain.lower()
        if model.mode == DomainType.direct and model.cdn_ip:
            raise ValidationError(f"Specifying CDN IP is only valid for CDN mode")

        if model.mode == DomainType.fake and not model.cdn_ip:
            model.cdn_ip = str(ipv4_list[0])

        # if model.mode==DomainType.fake and model.cdn_ip!=myip:
        #     raise ValidationError(f"Specifying CDN IP is only valid for CDN mode")

        # # Update the many-to-many relationship
        if len(model.show_domains) == Domain.query.count():
            model.show_domains = []

        if model.mode == DomainType.reality:
            model.servernames = (model.servernames or model.domain).lower()
            for v in set([model.domain, model.servernames]):
                for d in v.split(","):
                    if not d:
                        continue
                    if not hutils.network.is_domain_reality_friendly(d):
                        raise ValidationError(_("Domain is not REALITY friendly!") + f' {d}')

                    if not hutils.network.is_in_same_asn(d, ipv4_list[0]):
                        server_asn = hutils.network.get_ip_asn_name(ipv4_list[0])
                        domain_asn = hutils.network.get_ip_asn_name(dip)  # type: ignore
                        msg = _("selected domain for REALITY is not in the same ASN. To better use of the protocol, it is better to find a domain in the same ASN.") + \
                            (f"<br> Server ASN={server_asn}<br>{d}_ASN={domain_asn}" if server_asn or domain_asn else "")
                        hutils.flask.flash(msg, 'warning')

            for d in model.servernames.split(","):
                if not hutils.network.fallback_domain_compatible_with_servernames(model.domain, d):
                    raise ValidationError(_("REALITY Fallback domain is not compaitble with server names!") + f' {d} != {model.domain}')

        if (model.cdn_ip):
            try:
                hutils.network.auto_ip_selector.get_clean_ip(str(model.cdn_ip))
            except BaseException:
                raise ValidationError(_("Error in auto cdn format"))

        old_db_domain = get_domain(model.domain)
        if is_created or not old_db_domain or old_db_domain.mode != model.mode:
            # return hiddify.reinstall_action(complete_install=False, domain_changed=True)
            hutils.flask.flash_config_success(restart_mode=ApplyMode.apply, domain_changed=True)

    # def after_model_change(self,form, model, is_created):
    #     if model.show_domains.count==0:
    #         db.session.bulk_save_objects(ShowDomain(model.id,model.id))

    def on_model_delete(self, model):
        if len(Domain.query.all()) <= 1:
            raise ValidationError(f"at least one domain should exist")
        # ShowDomain.query.filter_by(related_id == model.id).delete()
        model.showed_by_domains = []
        # db.session.commit()
        hutils.flask.flash_config_success(restart_mode=ApplyMode.apply, domain_changed=True)

    def after_model_delete(self, model):
        if hutils.node.is_child():
            if not hutils.node.child.sync_with_parent():
                hutils.flask.flash(_('child.sync-failed'), 'danger')  # type: ignore

    def after_model_change(self, form, model, is_created):
        if hconfig(ConfigEnum.first_setup):
            set_hconfig(ConfigEnum.first_setup, False)
        if model.need_valid_ssl:
            commander(Command.get_cert, domain=model.domain)
        if hutils.node.is_child():
            if not hutils.node.child.sync_with_parent():
                hutils.flask.flash(_('child.sync-failed'), 'danger')  # type: ignore

    def is_accessible(self):
        if login_required(roles={Role.super_admin, Role.admin})(lambda: True)() != True:
            return False
        return True

    # def form_choices(self, field, *args, **kwargs):
    #     if field.type == "Enum":
    #         return [(enum_value.name, _(enum_value.name)) for enum_value in field.type.__members__.values()]
    #     return super().form_choices(field, *args, **kwargs)

    # @property
    # def server_ips(self):
    #     return hiddify.get_ip(4)

    def get_query(self):
        query = super().get_query()
        return query.filter(Domain.child_id == Child.current().id)




from urllib.parse import (
    urlencode, unquote, urlparse, parse_qsl, ParseResult
)
import webbrowser
from flask import g, request, render_template
from sys import version as python_version
from platform import platform
from json import dumps
from flask import g, request, render_template

import hiddifypanel
from hiddifypanel.models.config import hconfig
from hiddifypanel.models.config_enum import ConfigEnum


class __IssueUrl:
    def __init__(self, options):

        repoUrl = None

        self.opts = options

        if "repoUrl" in self.opts:
            repoUrl = self.opts["repoUrl"]

            try:
                del self.opts["user"]
                del self.opts["repo"]
            except BaseException:
                pass

        elif "user" in self.opts and "repo" in options:
            if repoUrl in self.opts:
                del self.opts["repoUrl"]

            repoUrl = "https://github.com/{0}/{1}".format(self.opts["user"], self.opts["repo"])
        else:
            raise KeyError('You need to specify either the `repoUrl` option or both the `user` and `repo` options')

        self.url = "{0}/issues/new".format(repoUrl)

        self.types = [
            'body', 'title', 'labels', 'template', 'milestone', 'assignee', 'projects'
        ]

    def add_url_params(self, url, params):
        """ Add GET params to provided URL being aware of existing.

        :param url: string of target URL
        :param params: dict containing requested params to be added
        :return: string with updated URL

        >> url = 'http://stackoverflow.com/test?answers=true'
        >> new_params = {'answers': False, 'data': ['some','values']}
        >> add_url_params(url, new_params)
        'http://stackoverflow.com/test?data=some&data=values&answers=false'

        Source: https://stackoverflow.com/a/25580545/3821823
        """
        # Unquoting URL first so we don't loose existing args
        url = unquote(url)
        # Extracting url info
        parsed_url = urlparse(url)
        # Extracting URL arguments from parsed URL
        get_args = parsed_url.query
        # Converting URL arguments to dict
        parsed_get_args = dict(parse_qsl(get_args))
        # Merging URL arguments dict with new params
        parsed_get_args.update(params)

        # Bool and Dict values should be converted to json-friendly values
        # you may throw this part away if you don't like it :)
        parsed_get_args.update(
            {k: dumps(v) for k, v in parsed_get_args.items()
             if isinstance(v, (bool, dict))}
        )

        # Converting URL argument to proper query string
        encoded_get_args = urlencode(parsed_get_args, doseq=True)
        # Creating new parsed result object based on provided with new
        # URL arguments. Same thing happens inside of urlparse.
        new_url = ParseResult(
            parsed_url.scheme, parsed_url.netloc, parsed_url.path,
            parsed_url.params, encoded_get_args, parsed_url.fragment
        ).geturl()

        return new_url

    def get_url(self):

        url = self.url

        for type in self.types:
            value = self.opts.get(type)
            if value is None:
                continue
            if type == "labels" or type == "projects":
                if not isinstance(value, list):
                    err = "The {0} option should be an array".format(type)
                    raise TypeError(err)

                value = ",".join(map(str, value))
                self.opts[type] = value

        self.url = self.add_url_params(url, self.opts)

        return self.url

    def opn(self):
        webbrowser.open(self.get_url(), 1)

# region private functions


def __generate_github_issue_link(title: str, issue_body: str) -> str:
    opts = {
        "user": 'hiddify',
        "repo": 'Hiddify-Manager',
        "title": title,
        "body": issue_body,
    }
    issue_link = str(__IssueUrl(opts).get_url())
    return str(issue_link)


def __github_issue_details() -> dict:
    details = {
        'hiddify_version': f'{hiddifypanel.__version__}',
        'python_version': f'{python_version}',
        'os_details': f'{platform()}',
        'user_agent': request.user_agent
    }
    return details


def __remove_sensetive_data_from_github_issue_link(issue_link: str):
    from hiddifypanel.auth import current_account
    if current_account.uuid:
        issue_link.replace(f'{current_account.uuid}', '*******************')

    issue_link.replace(request.host, '**********')
    issue_link.replace(hconfig(ConfigEnum.proxy_path), '**********')
    issue_link.replace(hconfig(ConfigEnum.proxy_path_admin), '**********')
    issue_link.replace(hconfig(ConfigEnum.proxy_path_client), '**********')


def __remove_unrelated_traceback_details(stacktrace: str) -> str:
    lines = stacktrace.splitlines()
    if len(lines) < 1:
        return ""

    output = ''
    skip_next_line = False
    for i, line in enumerate(lines):
        if i == 0:
            output += line + '\n'
            continue
        if skip_next_line:
            skip_next_line = False
            continue
        if line.strip().startswith('File'):
            if 'hiddify' in line.lower():
                output += line + '\n'
                if len(lines) > i + 1:
                    output += lines[i + 1] + '\n'
            skip_next_line = True

    return output

# endregion


def generate_github_issue_link_for_500_error(error, traceback: str, remove_sensetive_data: bool = True, remove_unrelated_traceback_datails: bool = True) -> str:

    if remove_unrelated_traceback_datails:
        traceback = __remove_unrelated_traceback_details(traceback)

    issue_details = __github_issue_details()

    issue_body = render_template('github_issue_body.j2', issue_details=issue_details, error=error, traceback=traceback)

    # Create github issue link
    issue_link = __generate_github_issue_link(f"Internal server error: {error.name if hasattr(error,'name') and error.name != None and error.name else 'Unknown'}", issue_body)

    if remove_sensetive_data:
        __remove_sensetive_data_from_github_issue_link(issue_link)

    return issue_link


def generate_github_issue_link_for_admin_sidebar() -> str:

    issue_body = render_template('github_issue_body.j2', issue_details=__github_issue_details())

    # Create github issue link
    issue_link = __generate_github_issue_link('Please fill the title properly', issue_body)
    return issue_link



import json
import copy
from flask import render_template, g
from hiddifypanel import hutils
from hiddifypanel.models import ProxyTransport, ProxyL3, ProxyProto, Domain
from flask_babel import gettext as _
from .xray import is_muxable_agent
OUTBOUND_LEVEL = 8


def configs_as_json(domains: list[Domain], remarks: str) -> str:
    '''Returns xray configs as json'''
    outbounds = []

    # TODO: check what are unsupported protocols in other apps
    unsupported_protos = {}
    if g.user_agent.get('is_v2rayng'):
        # TODO: ensure which protocols are not supported in v2rayng
        unsupported_protos = {ProxyProto.wireguard, ProxyProto.hysteria, ProxyProto.hysteria2, ProxyProto.tuic, ProxyProto.ss, ProxyProto.ssr, ProxyProto.ssh}

    for proxy in hutils.proxy.get_valid_proxies(domains):
        if unsupported_protos and proxy['proto'] in unsupported_protos:
            continue
        outbound = to_xray(proxy)
        if 'msg' not in outbound:
            outbounds.append(outbound)

    outbounds_len = len(outbounds)
    # reutrn no outbound
    if outbounds_len < 1:
        return ''

    all_configs = []
    base_config = json.loads(render_template('base_xray_config.json.j2', remarks=remarks))
    # multiple outbounds needs multiple whole base config not just one with multiple outbounds (at least for v2rayng)
    # https://github.com/2dust/v2rayNG/pull/2827#issue-2127534078
    if outbounds_len > 1:
        for out in outbounds:
            base_config['remarks'] = out['tag']
            base_config['outbounds'].insert(0, out)
            all_configs.append(copy.deepcopy(base_config))
            del base_config['outbounds'][0]
    else:  # single outbound
        base_config['outbounds'].insert(0, outbounds[0])
        all_configs = base_config

    json_configs = json.dumps(all_configs, indent=2, cls=hutils.proxy.ProxyJsonEncoder)
    return json_configs


def to_xray(proxy: dict) -> dict:
    outbound = {
        'tag': f'{proxy["extra_info"]} {proxy["name"]} § {proxy["port"]} {proxy["dbdomain"].id}',
        'protocol': str(proxy['proto']),
        'settings': {},
        'streamSettings': {},
        'mux': {  # default value
            'enabled': False,
            'concurrency': -1
        }
    }
    # add multiplex to outbound
    add_multiplex(outbound, proxy)

    # add stream setting to outbound
    add_stream_settings(outbound, proxy)

    # add protocol settings to outbound
    add_proto_settings(outbound, proxy)

    return outbound

# region proto settings


def add_proto_settings(base: dict, proxy: dict):
    if proxy['proto'] == ProxyProto.wireguard:
        add_wireguard_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.ss:
        add_shadowsocks_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.vless:
        add_vless_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.vmess:
        add_vmess_settings(base, proxy)
    elif proxy['proto'] == ProxyProto.trojan:
        proxy['password'] = proxy['uuid']
        add_trojan_settings(base, proxy)


def add_wireguard_settings(base: dict, proxy: dict):

    base['settings']['secretKey'] = proxy['wg_pk']
    base['settings']['reversed'] = [0, 0, 0]
    base['settings']['mtu'] = 1380  # optional
    base['settings']['peers'] = [{
        'endpoint': f'{proxy["server"]}:{int(proxy["port"])}',
        'publicKey': proxy["wg_server_pub"]
        # 'allowedIPs':'', 'preSharedKey':'', 'keepAlive':'' # optionals
    }]

    # optionals
    # base['settings']['address'] = [f'{proxy["wg_ipv4"]}/32',f'{proxy["wg_ipv6"]}/128']
    # base['settings']['workers'] = 4
    # base['settings']['domainStrategy'] = 'ForceIP' # default


def add_vless_settings(base: dict, proxy: dict):
    base['settings']['vnext'] = [
        {
            'address': proxy['server'],
            'port': proxy['port'],
            "users": [
                {
                    'id': proxy['uuid'],
                    'encryption': 'none',
                    # 'security': 'auto',
                    'flow': 'xtls-rprx-vision' if (proxy['transport'] == ProxyTransport.XTLS or base['streamSettings']['security'] == 'reality') else '',
                    'level': OUTBOUND_LEVEL
                }
            ]
        }
    ]


def add_vmess_settings(base: dict, proxy: dict):
    base['settings']['vnext'] = [
        {
            "address": proxy['server'],
            "port": proxy['port'],
            "users": [
                {
                    "id": proxy['uuid'],
                    "security": proxy['cipher'],
                    "level": OUTBOUND_LEVEL
                }
            ]
        }
    ]


def add_trojan_settings(base: dict, proxy: dict):
    base['settings']['servers'] = [
        {
            # 'email': proxy['uuid'], optional
            'address': proxy['server'],
            'port': proxy['port'],
            'password': proxy['password'],
            'level': OUTBOUND_LEVEL
        }
    ]


def add_shadowsocks_settings(base: dict, proxy: dict):
    base['settings']['servers'] = [
        {
            'address': proxy['server'],
            'port': proxy['port'],
            'method': proxy['cipher'],
            'password': proxy['password'],
            'uot': True,
            'level': OUTBOUND_LEVEL
            # 'email': '', optional
        }
    ]

# endregion


# region stream settings

def add_stream_settings(base: dict, proxy: dict):
    ss = base['streamSettings']
    ss['security'] = 'none'  # default

    # security
    if proxy['l3'] == ProxyL3.reality:
        ss['security'] = 'reality'
    elif proxy['l3'] in [ProxyL3.tls, ProxyL3.tls_h2, ProxyL3.tls_h2_h1]:
        ss['security'] = 'tls'

    # network and transport settings
    if ss['security'] == 'tls' or 'xtls':
        ss['tlsSettings'] = {
            'serverName': proxy['sni'],
            'allowInsecure': proxy['allow_insecure'],
            'fingerprint': proxy['fingerprint'],
            'alpn': [proxy['alpn']],
            # 'minVersion': '1.2',
            # 'disableSystemRoot': '',
            # 'enableSessionResumption': '',
            # 'pinnedPeerCertificateChainSha256': '',
            # 'certificates': '',
            # 'maxVersion': '1.3', # Go lang sets
            # 'cipherSuites': '', # Go lang sets
            # 'rejectUnknownSni': '', # default is false
        }
    if ss['security'] == 'reality':
        ss['network'] = proxy['transport']
        add_reality_stream(ss, proxy)
    if proxy['l3'] == ProxyL3.kcp:
        ss['network'] = 'kcp'
        add_kcp_stream(ss, proxy)

    if proxy['l3'] == ProxyL3.h3_quic:
        add_quic_stream(ss, proxy)

    if proxy['transport'] == 'tcp' or ss['security'] == 'reality' or (ss['security'] == 'none' and proxy['transport'] not in [ProxyTransport.httpupgrade, ProxyTransport.WS]):
        ss['network'] = proxy['transport']
        add_tcp_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.h2 and ss['security'] == 'none' and ss['security'] != 'reality':
        ss['network'] = proxy['transport']
        add_http_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.grpc:
        ss['network'] = proxy['transport']
        add_grpc_stream(ss, proxy)
    if proxy['transport'] == ProxyTransport.httpupgrade:
        ss['network'] = proxy['transport']
        add_httpupgrade_stream(ss, proxy)
    if proxy['transport'] == 'ws':
        ss['network'] = proxy['transport']
        add_ws_stream(ss, proxy)

    # tls fragmentaion
    add_tls_fragmentation_stream_settings(base, proxy)


def add_tcp_stream(ss: dict, proxy: dict):
    if proxy['l3'] == ProxyL3.http:
        ss['tcpSettings'] = {
            'header': {
                'type': 'http',
                'request': {
                    'path': [proxy['path']]
                }
            }
            # 'acceptProxyProtocol': False
        }
    else:
        ss['tcpSettings'] = {
            'header': {
                'type': 'none'
            }
            # 'acceptProxyProtocol': False
        }


def add_http_stream(ss: dict, proxy: dict):
    ss['httpSettings'] = {
        'host': proxy['host'],
        'path': proxy['path'],
        # 'read_idle_timeout': 10,  # default disabled
        # 'health_check_timeout': 15,  # default is 15
        # 'method': 'PUT',  # default is 15
        # 'headers': {

        # }
    }


def add_ws_stream(ss: dict, proxy: dict):
    ss['wsSettings'] = {
        'path': proxy['path'],
        'headers': {
            "Host": proxy['host']
        }
        # 'acceptProxyProtocol': False,
    }


def add_grpc_stream(ss: dict, proxy: dict):
    ss['grpcSettings'] = {
        'serviceName': proxy['path'],  # proxy['path'] is equal toproxy['grpc_service_name']
        'idle_timeout': 115,  # by default, the health check is not enabled. may solve some "connection drop" issues
        'health_check_timeout': 20,  # default is 20
        # 'initial_windows_size': 0,  # 0 means disabled. greater than 65535 means Dynamic Window mechanism will be disabled
        # 'permit_without_stream': False, # health check performed when there are no sub-connections
        # 'multiMode': false, # experimental
    }


def add_httpupgrade_stream(ss: dict, proxy: dict):
    ss['httpupgradeSettings'] = {
        'path': proxy['path'],
        'host': proxy['host'],
        # 'acceptProxyProtocol': '', for inbounds only
    }


def add_kcp_stream(ss: dict, proxy: dict):
    # TODO: fix server side configs first
    ss['kcpSettings'] = {}
    return
    ss['kcpSettings'] = {
        'seed': proxy['path'],
        # 'mtu': 1350, # optional, default value is written
        # 'tti': 50, # optional, default value is written
        # 'uplinkCapacity': 5, # optional, default value is written
        # 'downlinkCapacity': 20, # optional, default value is written
        # 'congestion':False, # optional, default value is written
        # 'readBufferSize': 2,# optional, default value is written
        # 'writeBufferSize':2 # optional, default value is written
        # 'header': { # must be same as server (hiddify doesn't use yet)
        #     'type': 'none'  # choices: none(default), srtp, utp, wechat-video, dtls, wireguards
        # }
    }


def add_quic_stream(ss: dict, proxy: dict):
    # TODO: fix server side configs first
    ss['quicSettings'] = {}
    return
    ss['quicSettings'] = {
        'security': 'chacha20-poly1305',
        'key': proxy['path'],
        'header': {
            'type': 'none'
        }
    }


def add_reality_stream(ss: dict, proxy: dict):
    ss['realitySettings'] = {
        'serverName': proxy['sni'],
        'fingerprint': proxy['fingerprint'],
        'shortId': proxy['reality_short_id'],
        'publicKey': proxy['reality_pbk'],
        'show': False,
    }


def add_tls_fragmentation_stream_settings(base: dict, proxy: dict):
    '''Adds tls fragment in the outbounds if tls fragmentation is enabled'''
    if base['streamSettings']['security'] in ['tls', 'reality']:
        if proxy.get('tls_fragment_enable'):
            base['streamSettings']['sockopt'] = {
                'dialerProxy': 'fragment',
                'tcpKeepAliveIdle': 100,
                'tcpNoDelay': True,  # recommended to be enabled with "tcpMptcp": true.
                "mark": 255
                # 'tcpFastOpen': True, # the system default setting be used.
                # 'tcpKeepAliveInterval': 0, # 0 means default GO lang settings, -1 means not enable
                # 'tcpcongestion': bbr, # Not configuring means using the system default value
                # 'tcpMptcp': True, # need to be enabled in both server and client configuration (not supported by panel yet)
            }

# endregion


def add_multiplex(base: dict, proxy: dict):
    if proxy.get('mux_enable') != "xray":
        return

    concurrency = proxy['mux_max_connections']
    if concurrency and concurrency > 0:
        base['mux']['enabled'] = True
        base['mux']['concurrency'] = concurrency
        base['mux']['xudpConcurrency'] = concurrency
        base['mux']['xudpProxyUDP443'] = 'reject'


def add_tls_tricks_to_dict(d: dict, proxy: dict):
    if proxy.get('tls_fragment_enable'):
        if g.user_agent.get('is_shadowrocket'):
            d['fragment'] = f'1,{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]}'
        else:
            d['fragment'] = f'{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]},tlshello'

    if proxy.get("tls_mixed_case"):
        d['mc'] = 1
    if proxy.get("tls_padding_enable"):
        d['padsize'] = proxy["tls_padding_length"]


def add_mux_to_dict(d: dict, proxy):
    if not is_muxable_agent(proxy):
        return

    # according to github.com/hiddify/ray2sing/
    d['muxtype'] = proxy["mux_protocol"]
    d['muxmaxc'] = proxy["mux_max_connections"]
    d['mux'] = proxy['mux_min_streams']
    d['muxsmax'] = proxy["mux_max_streams"]
    d['muxpad'] = proxy["mux_padding_enable"]

    if proxy.get('mux_brutal_enable'):
        d['muxup'] = proxy["mux_brutal_up_mbps"]
        d['muxdown'] = proxy["mux_brutal_down_mbps"]


def add_tls_tricks_to_link(proxy: dict) -> str:
    out = {}
    add_tls_tricks_to_dict(out, proxy)
    return hutils.encode.convert_dict_to_url(out)


def add_mux_to_link(proxy: dict) -> str:
    out = {}
    add_mux_to_dict(out, proxy)
    return hutils.encode.convert_dict_to_url(out)



import datetime
import json
import copy
from flask import render_template, request, g
from hiddifypanel import hutils
from hiddifypanel.models import Proxy, ProxyTransport, ProxyL3, ProxyCDN, ProxyProto, Domain, ConfigEnum, DomainType, hconfig
from flask_babel import gettext as _

OUTBOUND_LEVEL = 8


def is_muxable_agent(proxy: dict) -> bool:
    if not proxy.get('mux_enable'):
        return False
    if proxy.get('mux_enable') == "xray" and g.user_agent.get('is_singbox'):
        return False
    if proxy.get('mux_enable') == "singbox" and not g.user_agent.get('is_singbox'):
        return False
    return True


def to_link(proxy: dict) -> str | dict:
    if 'error' in proxy:
        return proxy

    orig_name_link = (proxy['extra_info'] + " " + proxy["name"]).strip()
    name_link = hutils.encode.url_encode(orig_name_link)
    if proxy['proto'] == 'vmess':
        # print(proxy)
        vmess_type = None
        if proxy["transport"] == 'tcp':
            vmess_type = 'http'
        if 'grpc_mode' in proxy:
            vmess_type = proxy['grpc_mode']
        vmess_data = {"v": "2",
                      "ps": orig_name_link,
                      "add": proxy['server'],
                      "port": proxy['port'],
                      "id": proxy["uuid"],
                      "aid": 0,
                      "scy": proxy['cipher'],
                      "net": proxy["transport"],
                      "type": vmess_type or "none",
                      "host": proxy.get("host", ""),
                      "alpn": proxy.get("alpn", "h2,http/1.1"),
                      "path": proxy["path"] if "path" in proxy else "",
                      "tls": "tls" if "tls" in proxy["l3"] else "",
                      "sni": proxy["sni"],
                      "fp": proxy["fingerprint"]
                      }

        if 'reality' in proxy["l3"]:
            vmess_data['tls'] = "reality"
            vmess_data['pbk'] = proxy['reality_pbk']
            vmess_data['sid'] = proxy['reality_short_id']

        add_tls_tricks_to_dict(vmess_data, proxy)
        add_mux_to_dict(vmess_data, proxy)

        return "vmess://" + hutils.encode.do_base_64(f'{json.dumps(vmess_data,cls=hutils.proxy.ProxyJsonEncoder)}')
    if proxy['proto'] == 'ssh':
        baseurl = 'ssh://'
        if g.user_agent.get('is_streisand'):
            streisand_ssh = hutils.encode.do_base_64(f'{proxy["uuid"]}:0:{proxy["private_key"]}::@{proxy["server"]}:{proxy["port"]}')
            baseurl += f'{streisand_ssh}#{name_link}'
        else:
            hk = ",".join(proxy["host_key"])
            pk = proxy["private_key"].replace('\n', '')
            baseurl += f'{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}/?file=ssh&pk={pk}&hk={hk}#{name_link}'

        return baseurl
    if proxy['proto'] == "ssr":
        baseurl = f'ssr://{proxy["cipher"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}'
        return baseurl
    if proxy['proto'] in ['ss', 'v2ray']:
        baseurl = f'ss://{proxy["cipher"]}:{proxy["password"]}@{proxy["server"]}:{proxy["port"]}'
        if proxy['mode'] == 'faketls':
            return f'{baseurl}?plugin=obfs-local%3Bobfs%3Dtls%3Bobfs-host%3D{proxy["fakedomain"]}%3Budp-over-tcp=true#{name_link}'
        # if proxy['mode'] == 'shadowtls':
        #     return f'{baseurl}?plugin=shadow-tls%3Bpassword%3D{proxy["proxy_path"]}%3Bhost%3D{proxy["fakedomain"]}%3Budp-over-tcp=true#{name_link}'
        if proxy['proto'] == 'v2ray':
            return f'{baseurl}?plugin=v2ray-plugin%3Bmode%3Dwebsocket%3Bpath%3D{proxy["path"]}%3Bhost%3D{proxy["host"]}%3Btls%3Budp-over-tcp=true#{name_link}'
        if proxy['transport'] == 'shadowsocks':
            return baseurl
    if proxy['proto'] == 'tuic':
        baseurl = f'tuic://{proxy["uuid"]}:{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?congestion_control=cubic&udp_relay_mode=native&sni={proxy["sni"]}&alpn=h3'
        if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
            baseurl += "&allow_insecure=1"
        return f"{baseurl}#{name_link}"
    if proxy['proto'] == 'hysteria2':
        baseurl = f'hysteria2://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?hiddify=1&obfs=salamander&obfs-password={proxy["hysteria_obfs_password"]}&sni={proxy["sni"]}'
        if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
            baseurl += "&insecure=1"
        return f"{baseurl}#{name_link}"
    if proxy['proto'] == ProxyProto.wireguard:
        if g.user_agent.get('is_streisand'):
            return f'wireguard://{proxy["server"]}:{proxy["port"]}?private_key={proxy["wg_pk"]}&peer_public_key={proxy["wg_server_pub"]}&pre_shared_key={proxy["wg_psk"]}&reserved=0,0,0#{name_link}'
        else:
            # hiddify_format =
            # f'wg://{proxy["server"]}:{proxy["port"]}/?pk={proxy["wg_pk"]}&local_address={proxy["wg_ipv4"]}/32&peer_pk={proxy["wg_server_pub"]}&pre_shared_key={proxy["wg_psk"]}&workers=4&mtu=1380&reserved=0,0,0&ifp={proxy["wg_noise_trick"]}#{name_link}'
            return f'wg://{proxy["server"]}:{proxy["port"]}?publicKey={proxy["wg_pub"]}&privateKey={proxy["wg_pk"]}=&presharedKey={proxy["wg_psk"]}&ip=10.0.0.1&mtu=1380&keepalive=30&udp=1&reserved=0,0,0&ifp={proxy["wg_noise_trick"]}#{name_link}'

    baseurl = f'{proxy["proto"]}://{proxy["uuid"]}@{proxy["server"]}:{proxy["port"]}?hiddify=1'
    baseurl += f'&sni={proxy["sni"]}&type={proxy["transport"]}'
    baseurl += f"&alpn={proxy['alpn']}"

    # the ray2sing supports vless, vmess and trojan tls tricks and mux
    # the vmess handled already

    baseurl += add_mux_to_link(proxy)
    baseurl += add_tls_tricks_to_link(proxy)

    # infos+=f'&alpn={proxy["alpn"]}'
    baseurl += f'&path={proxy["path"]}' if "path" in proxy else ""
    baseurl += f'&host={proxy["host"]}' if "host" in proxy else ""
    if "grpc" == proxy["transport"]:
        baseurl += f'&serviceName={proxy["grpc_service_name"]}&mode={proxy["grpc_mode"]}'
    # print(proxy['cdn'],proxy["transport"])
    if request.args.get("fragment"):
        baseurl += f'&fragment=' + request.args.get("fragment")  # type: ignore
    if "ws" == proxy["transport"] and proxy['cdn'] and request.args.get("fragment_v1"):
        baseurl += f'&fragment_v1=' + request.args.get("fragment_v1")  # type: ignore
    if 'vless' == proxy['proto']:
        baseurl += "&encryption=none"
    if proxy.get('fingerprint', 'none') != 'none':
        baseurl += "&fp=" + proxy['fingerprint']
    if proxy['l3'] != 'quic':
        if proxy.get('l3') != ProxyL3.reality and (proxy.get('transport') == ProxyTransport.tcp or proxy.get('transport') == ProxyTransport.httpupgrade) and proxy['proto'] in [ProxyProto.vless, ProxyProto.trojan]:
            baseurl += '&headerType=http'
        else:
            baseurl += '&headerType=None'
    if proxy['mode'] == 'Fake' or proxy['allow_insecure']:
        baseurl += "&allowInsecure=true"
    if proxy.get('flow'):
        baseurl += f'&flow={proxy["flow"]}'

    infos = f'#{name_link}'

    if 'reality' in proxy["l3"]:
        return f"{baseurl}&security=reality&pbk={proxy['reality_pbk']}&sid={proxy['reality_short_id']}{infos}"
    if 'tls' in proxy['l3']:
        return f'{baseurl}&security=tls{infos}'
    if proxy['l3'] == 'http':
        return f'{baseurl}&security=none{infos}'
    return proxy


def make_v2ray_configs(user, user_activate, domains: list[Domain], expire_days, ip_debug, db_domain, has_auto_cdn, asn, profile_title, **kwargs) -> str:
    res = []

    ua = hutils.flask.get_user_agent()
    if hconfig(ConfigEnum.show_usage_in_sublink):

        if not ua['is_hiddify']:

            fake_ip_for_sub_link = datetime.datetime.now().strftime(f"%H.%M--%Y.%m.%d.time:%H%M")
            # if ua['app'] == "Fair1":
            #     res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{round(user.current_usage_GB,3)}/{user.usage_limit_GB}GB_Remain:{expire_days}days')
            # else:

            # res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hutils.encode.url_encode(profile_title)}')

            name = '⏳' if user_activate else '✖'
            if user.usage_limit_GB < 1000:
                name += f'{round(user.current_usage_GB,3)}/{str(user.usage_limit_GB).replace(".0","")}GB'
            elif user.usage_limit_GB < 100000:
                name += f'{round(user.current_usage_GB/1000,3)}/{str(round(user.usage_limit_GB/1000,1)).replace(".0","")}TB'
            else:
                res.append("#No Usage Limit")
            if expire_days < 1000:
                name += " " + _(f'📅%(expire_days)s days', expire_days=expire_days)
            else:
                res.append("#No Time Limit")

            name = name.strip()
            if len(name) > 3:
                res.append(f'trojan://1@{fake_ip_for_sub_link}?sni=fake_ip_for_sub_link&security=tls#{hutils.encode.url_encode(name)}')

    if ua['is_browser']:
        res.append(f'#Hiddify auto ip: {ip_debug}')

    if not user_activate:

        if hconfig(ConfigEnum.lang) == 'fa':
            res.append('trojan://1@1.1.1.1#' + hutils.encode.url_encode('✖بسته شما به پایان رسید'))
        else:
            res.append('trojan://1@1.1.1.1#' + hutils.encode.url_encode('✖Package_Ended'))
        return "\n".join(res)

    for pinfo in hutils.proxy.get_valid_proxies(domains):
        link = to_link(pinfo)
        if 'msg' not in link:
            res.append(link)
    return "\n".join(res)


def add_tls_tricks_to_dict(d: dict, proxy: dict):
    if proxy.get('tls_fragment_enable'):
        if g.user_agent.get('is_shadowrocket'):
            d['fragment'] = f'1,{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]}'
        else:
            d['fragment'] = f'{proxy["tls_fragment_size"]},{proxy["tls_fragment_sleep"]},tlshello'

    if proxy.get("tls_mixed_case"):
        d['mc'] = 1
    if proxy.get("tls_padding_enable"):
        d['padsize'] = proxy["tls_padding_length"]


def add_mux_to_dict(d: dict, proxy):
    if not is_muxable_agent(proxy):
        return

    # according to github.com/hiddify/ray2sing/
    d['muxtype'] = proxy["mux_protocol"]
    d['muxmaxc'] = proxy["mux_max_connections"]
    d['mux'] = proxy['mux_min_streams']
    d['muxsmax'] = proxy["mux_max_streams"]
    d['muxpad'] = proxy["mux_padding_enable"]

    if proxy.get('mux_brutal_enable'):
        d['muxup'] = proxy["mux_brutal_up_mbps"]
        d['muxdown'] = proxy["mux_brutal_down_mbps"]


def add_tls_tricks_to_link(proxy: dict) -> str:
    out = {}
    add_tls_tricks_to_dict(out, proxy)
    return hutils.encode.convert_dict_to_url(out)


def add_mux_to_link(proxy: dict) -> str:
    out = {}
    add_mux_to_dict(out, proxy)
    return hutils.encode.convert_dict_to_url(out)



import base64
from typing import Tuple, Any
from uuid import UUID


def is_uuid_valid(uuid: str, version: int = 4) -> bool:
    try:
        uuid_obj = UUID(uuid, version=version)
    except ValueError:
        return False
    return str(uuid_obj) == uuid


def get_uuid_from_url_path(path: str, section_index: int = 2) -> str | None:
    """
    Takes a URL path and extracts the UUID at the specified section index.

    Args:
        path (str): The URL path from which to extract the UUID.
        section_index (int, optional): The index of the section in the URL path where the UUID is located. Defaults to 2, because in past the UUID was in the second section of path of url.

    Returns:
        str | None: The extracted UUID as a string if found, or None if not found.
    """
    s_index = 1
    for section in path.lstrip('/').split('/'):
        if is_uuid_valid(section, 4):
            if s_index == section_index:
                return section
        s_index += 1
    return None


def parse_login_id(raw_id: str) -> Tuple[Any | None, str | None]:
    """
    Parses the given raw ID to extract the account type and ID.
    Args:
        raw_id (str): The raw ID to be parsed.
    Returns:
        Tuple[Any | None, str | None]: A tuple containing the account type and ID.
            The account type is either AccountType.admin or AccountType.user
            and the ID is a string. If the raw ID cannot be parsed, None is returned
            for both the account type and ID.
    """
    splitted = raw_id.split('_')
    if len(splitted) < 2:
        return None, None
    admin_or_user, id = splitted
    from hiddifypanel.models.role import AccountType
    account_type = AccountType.admin if admin_or_user == 'admin' else AccountType.user
    if not id or not account_type:
        return None, None
    return account_type, id


def add_basic_auth_to_url(url: str, username: str, password: str) -> str:
    if 'https://' in url:
        return url.replace('https://', f'https://{username}:{password}@')
    elif 'http://' in url:
        return url.replace('http://', f'http://{username}:{password}@')
    else:
        return url

# region unused(never mentioned in codebase)


# def is_uuid_in_url_path(path: str) -> bool:
#     for section in path.split('/'):
#         if is_uuid_valid(section, 4):
#             return True
#     return False


# def get_basic_auth_from_auth_header(auth_header: str) -> str | None:
#     if auth_header.startswith('Basic'):
#         return auth_header.split('Basic ')[1].strip()
#     return None


# def parse_basic_auth_header(auth_header: str) -> tuple[str, str] | None:
#     if not auth_header.startswith('Basic'):
#         return None
#     header_value = auth_header.split('Basic ')
#     if len(header_value) < 2:
#         return None
#     username, password = map(lambda item: item.strip(), base64.urlsafe_b64decode(header_value[1].strip()).decode('utf-8').split(':'))
#     return (username, password) if username and password else None
# endregion



from typing import Optional
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy_utils import UUIDType
import re
import os
from sqlalchemy import Row, text, Sequence


db: SQLAlchemy = SQLAlchemy()
db.UUID = UUIDType  # type: ignore


def init_app(app):
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = True
    db.init_app(app)


def db_execute(query: str, return_val: bool = False, commit: bool = False, **params: dict):
    q = db.session.execute(text(query), params)
    if commit:
        db.session.commit()
    if return_val:
        return q.fetchall()

    # with db.engine.connect() as connection:
    #     res = connection.execute(text(query), params)
    #     connection.commit()s
    # return res



#!/usr/bin/env python3
from flask_classful import FlaskView
from flask_babel import gettext as _
from flask import render_template

from hiddifypanel.models import *


class CommercialInfo(FlaskView):

    def index(self):
        return render_template('commercial_info.html')



from flask_apispec import FlaskApiSpec
from flask_apispec.extension import FlaskApiSpec
from marshmallow import Schema, fields


from apiflask import APIBlueprint
from flask_restful import Api
from .tgbot import bot, register_bot, TGBotResource
from . import tgbot
from .tgmsg import SendMsgResource
from .resources import *
bp = APIBlueprint("api_v1", __name__, url_prefix="/<proxy_path>/api/v1/", tag="api_v1", enable_openapi=False)
bp_uuid = APIBlueprint("api_v1_uuid", __name__, url_prefix="/<proxy_path>/<uuid:secret_uuid>/api/v1/", template_folder="templates", enable_openapi=False)
api = Api(bp)
api_uuid = Api(bp_uuid)


def init_app(app):
    tgbot.init_app(app)
    api.add_resource(UserResource, "/user/")
    api.add_resource(AdminUserResource, "/admin/")
    # api.add_resource(TGBotResource, "/tgbot/")
    api.add_resource(SendMsgResource, "/send_msg/")

    api_uuid.add_resource(UserResource, "/user/")
    api_uuid.add_resource(AdminUserResource, "/admin/")
    api_uuid.add_resource(TGBotResource, "/tgbot/")
    api_uuid.add_resource(SendMsgResource, "/send_msg/")

    # with app.app_context():
    #     register_bot()
    app.register_blueprint(bp)
    app.register_blueprint(bp_uuid)



from enum import auto
import uuid
from flask import g
from hiddifypanel.models.usage import DailyUsage
from sqlalchemy import event, Column, Integer, Enum, Boolean, ForeignKey
from strenum import StrEnum
from apiflask import abort
from flask_babel import gettext as __
from flask_babel import lazy_gettext as _
from hiddifypanel.database import db, db_execute
from hiddifypanel.models.role import Role
from hiddifypanel.models.base_account import BaseAccount
from sqlalchemy_serializer import SerializerMixin


class AdminMode(StrEnum):
    """
    The "UserMode" class is an enumeration that defines five possible modes: "no_reset", "monthly", "weekly",
    "daily", and "disable". These modes represent different settings that can be applied to a user account,
    such as the frequency at which data is reset or whether the account is currently disabled. The class is
    implemented using the "StrEnum" base class and the "auto()" function to generate unique values for each mode.
    """
    super_admin = auto()
    admin = auto()
    agent = auto()


class AdminUser(BaseAccount, SerializerMixin):
    """
    This is a model class for a user in a database that includes columns for their ID, UUID, name, online status,
    account expiration date, usage limit, package days, mode, start date, current usage, last reset time, and comment.
    """
    id = Column(Integer, primary_key=True, autoincrement=True)
    mode = Column(Enum(AdminMode), default=AdminMode.agent, nullable=False)
    can_add_admin = Column(Boolean, default=False, nullable=False)
    max_users = Column(Integer, default=100, nullable=False)
    max_active_users = Column(Integer, default=100, nullable=False)
    users = db.relationship('User', backref='admin')  # type: ignore
    usages = db.relationship('DailyUsage', backref='admin')  # type: ignore
    parent_admin_id = Column(Integer, ForeignKey('admin_user.id'), default=1)
    parent_admin = db.relationship('AdminUser', remote_side=[id], backref='sub_admins')  # type: ignore

    @property
    def role(self) -> Role | None:
        match self.mode:
            case AdminMode.super_admin:
                return Role.super_admin
            case AdminMode.admin:
                return Role.admin
            case AdminMode.agent:
                return Role.agent
        return None

    @staticmethod
    def form_schema(schema):
        return schema.dump(AdminUser())

    def to_schema(self):
        admin_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.admin.admin_user_api import AdminSchema
        return AdminSchema().load(admin_dict)

    def get_id(self) -> str | None:
        return f'admin_{self.id}'

    def to_dict(self, convert_date=True, dump_id=False) -> dict:
        base = super().to_dict()
        if dump_id:
            base['id'] = self.id
        from hiddifypanel.models import hconfig, ConfigEnum
        return {**base,
                'mode': self.mode,
                'can_add_admin': self.can_add_admin,
                'parent_admin_uuid': self.parent_admin.uuid if self.parent_admin else None,
                'lang': hconfig(ConfigEnum.admin_lang)
                }

    @classmethod
    def by_uuid(cls, uuid: str, create: bool = False) -> BaseAccount | None:
        account = AdminUser.query.filter(AdminUser.uuid == uuid).first()
        if not account and create:
            dbuser = AdminUser(uuid=uuid, name="unknown", parent_admin_id=AdminUser.current_admin_or_owner().id)
            db.session.add(dbuser)
            db.session.commit()
            account = AdminUser.by_uuid(uuid, False)

        return account

    @classmethod
    def add_or_update(cls, commit: bool = True, **data):

        dbuser = super().add_or_update(commit=commit, **data)

        if dbuser.id != 1:
            parent = data.get('parent_admin_uuid')
            if parent == data['uuid'] or not parent:
                parent_admin = cls.current_admin_or_owner()
            else:
                parent_admin = cls.by_uuid(parent, create=True)
            dbuser.parent_admin_id = parent_admin.id  # type: ignore

        dbuser.mode = data.get('mode', AdminMode.agent)
        dbuser.can_add_admin = data.get('can_add_admin') or False
        if commit:
            db.session.commit()
        return dbuser

    def recursive_users_query(self):
        from .user import User
        admin_ids = self.recursive_sub_admins_ids()
        return User.query.filter(User.added_by.in_(admin_ids))

    def can_have_more_users(self):
        if self.mode == AdminMode.super_admin:
            return True
        users_count = self.recursive_users_query().count()
        if self.max_users < users_count:
            return False
        if users_count <= self.max_active_users:
            return True

        actives = [u for u in self.recursive_users_query().all() if u.is_active]
        return len(actives) <= self.max_active_users

    def recursive_sub_admins_ids(self, depth=20, seen=None):
        if seen is None:
            seen = set()
        sub_admin_ids = []
        if self.id not in seen:
            sub_admin_ids.append(self.id)
            seen.add(self.id)
        if depth > 0:
            for sub_admin in self.sub_admins:
                sub_admin_ids += sub_admin.recursive_sub_admins_ids(depth - 1, seen=seen)
        return sub_admin_ids

    def remove(self):
        if self.id == 1 or self.id == g.account.id:
            # raise ValidationError(_("Owner can not be deleted!"))
            abort(422, __("Owner can not be deleted!"))
        users = self.recursive_users_query().all()
        for u in users:
            u.added_by = g.account.id

        DailyUsage.query.filter(DailyUsage.admin_id.in_(self.recursive_sub_admins_ids())).update({'admin_id': g.account.id})
        AdminUser.query.filter(AdminUser.id.in_(self.recursive_sub_admins_ids())).delete()

        db.session.commit()

    def __str__(self):
        return str(self.name)

    @staticmethod
    def get_super_admin():
        admin = AdminUser.by_id(1)
        if not admin:
            db.session.add(AdminUser(id=1, uuid=str(uuid.uuid4()), name="Owner", mode=AdminMode.super_admin, comment=""))
            db.session.commit()

            db_execute("update admin_user set id=1 where name='Owner'", commit=True)
            admin = AdminUser.by_id(1)

        return admin

    @staticmethod
    def get_super_admin_uuid() -> str:
        return AdminUser.get_super_admin().uuid

    @staticmethod
    def current_admin_or_owner():
        if g and hasattr(g, 'account') and g.account and isinstance(g.account, AdminUser):
            return g.account
        return AdminUser.query.filter(AdminUser.id == 1).first()


@event.listens_for(AdminUser, "before_insert")
def before_insert(mapper, connection, target):
    from hiddifypanel import hutils
    hutils.model.gen_username(target)
    hutils.model.gen_password(target)



from flask_login import LoginManager, current_user, user_accessed, user_logged_in, COOKIE_NAME, AUTH_HEADER_NAME, logout_user
from flask import g, redirect, request, session
from hiddifypanel.hutils.flask import hurl_for
from flask_login.utils import _get_user
from flask import current_app
from functools import wraps
from apiflask import abort

from hiddifypanel.models import AdminUser, User, Role, AccountType
import hiddifypanel.panel.hiddify as hiddify
from hiddifypanel import hutils


class CustumLoginManager(LoginManager):

    def _load_user(self):
        if hasattr(g, "account") and g.account:
            print(g.account)
            return g.account

        if self._user_callback is None and self._request_callback is None:
            raise Exception(
                "Missing user_loader or request_loader. Refer to "
                "http://flask-login.readthedocs.io/#how-it-works "
                "for more info."
            )

        user_accessed.send(current_app._get_current_object())  # type: ignore

        # Check SESSION_PROTECTION
        if self._session_protection_failed():
            return self._update_request_context_with_user()

        # user = self.header_auth()
        user = None

        account_id = ''
        # Load user from Flask Session
        if hutils.flask.is_api_call(request.path):
            if hutils.flask.is_user_api_call():
                account_id = session.get("_user_id")
            else:
                account_id = session.get("_admin_id")
        elif hutils.flask.is_user_panel_call():
            account_id = session.get("_user_id")
        else:
            account_id = session.get("_admin_id")

        if account_id is not None and self._user_callback is not None:
            user = self._user_callback(account_id)

        # Load user from Remember Me Cookie or Request Loader
        if user is None:
            config = current_app.config
            cookie_name = config.get("REMEMBER_COOKIE_NAME", COOKIE_NAME)
            header_name = config.get("AUTH_HEADER_NAME", AUTH_HEADER_NAME)
            has_cookie = (
                cookie_name in request.cookies and session.get("_remember") != "clear"
            )
            # if header_name in request.headers:
            #     header = request.headers[header_name]
            #     user = self._load_user_from_header(header)
            # if self._request_callback:
            #     user = self._load_user_from_request(request)
            if (cookie := request.cookies.get(cookie_name)):
                user = self._load_user_from_remember_cookie(cookie)

        return self._update_request_context_with_user(user)

    def header_auth(self) -> User | AdminUser | None:
        auth_header: str = request.headers.get("Authorization")
        print("DDDDDDDDDDDDD", auth_header)
        if not auth_header:
            return

        account = None
        is_api_call = False

        if hutils.flask.is_api_call(request.path):
            if apikey := hutils.auth.get_apikey_from_auth_header(auth_header):
                account = User.by_uuid(apikey) or AdminUser.by_uuid(apikey)
                is_api_call = True
        else:
            uname = request.authorization.username
            pword = request.authorization.password
            account = AdminUser.by_username_password(uname, pword) if hutils.flask.is_admin_proxy_path() else User.by_username_password(uname, pword)

        if account:
            g.account = account
            # g.account_uuid = account.uuid
            g.is_admin = hutils.flask.is_admin_role(account.role)  # type: ignore
            login_user(account)
        # else:
        #     print("logout")
        #     logout_user()

        return account


def login_user(user: AdminUser | User, remember=False, duration=None, force=False, fresh=True):
    g.account = user
    if not user.is_active:
        return False

    account_id = getattr(user, current_app.login_manager.id_attribute)()  # type: ignore
    if user.role in {Role.super_admin, Role.admin, Role.agent}:
        session["_admin_id"] = account_id
    else:
        session["_user_id"] = account_id
    session["_fresh"] = fresh
    session["_id"] = current_app.login_manager._session_identifier_generator()  # type: ignore

    # if remember:
    #     session["_remember"] = "set"
    #     if duration is not None:
    #         try:
    #             # equal to timedelta.total_seconds() but works with Python 2.6
    #             session["_remember_seconds"] = (
    #                 duration.microseconds
    #                 + (duration.seconds + duration.days * 24 * 3600) * 10**6
    #             ) / 10.0**6
    #         except AttributeError as e:
    #             raise Exception(
    #                 f"duration must be a datetime.timedelta, instead got: {duration}"
    #             ) from e

    current_app.login_manager._update_request_context_with_user(user)  # type: ignore
    user_logged_in.send(current_app._get_current_object(), user=_get_user())  # type: ignore
    return True


def login_required(roles: set[Role] | None = None):
    def wrapper(fn):
        @wraps(fn)
        def decorated_view(*args, **kwargs):
            if not current_user.is_authenticated:
                return auth.redirect_to_login()  # type: ignore
            if roles:
                account_role = current_user.role
                if account_role not in roles:
                    return auth.redirect_to_login()  # type: ignore
            return fn(*args, **kwargs)
        return decorated_view
    return wrapper


def get_account_by_api_key(api_key, is_admin):
    return AdminUser.by_uuid(api_key) if is_admin else User.by_uuid(api_key)


def get_account_by_uuid(uuid, is_admin):
    return AdminUser.by_uuid(uuid) if is_admin else User.by_uuid(uuid)


def init_app(app):
    # login_manager = LoginManager()
    login_manager = CustumLoginManager()
    login_manager.init_app(app)

    @app.before_request
    def auth_from_url():
        account = None

        is_admin_path = hutils.flask.is_admin_proxy_path()
        next_url = None
        print("--------1")
        if auth_header := request.headers.get("Authorization"):
            apikey = hutils.auth.get_apikey_from_auth_header(auth_header)
            account = get_account_by_api_key(apikey, is_admin_path)
            if not account:
                logout_user()

        if not account and (uuid := hutils.auth.get_uuid_from_url_path(request.path)):
            account = get_account_by_uuid(uuid, is_admin_path)
            if not account:
                logout_user()
            else:
                next_url = request.url.replace(f'/{uuid}/', '/' if is_admin_path else '/client/')
        print(request.authorization)
        if not account and request.authorization:
            uname = request.authorization.username
            pword = request.authorization.password
            account = AdminUser.by_username_password(uname, pword) if is_admin_path else User.by_username_password(uname, pword)
            print(request.authorization, account)
            if not account:
                logout_user()
        if account:
            g.account = account
            # g.account_uuid = account.uuid
            g.is_admin = hutils.flask.is_admin_role(account.role)  # type: ignore
            login_user(account, force=True)
            print("loggining in")
            if next_url is not None:
                if 0 and g.user_agent['is_browser']:
                    return redirect(next_url)
                else:
                    print(next_url)
                    request.url = next_url

    @login_manager.user_loader
    def cookie_auth(id: str) -> User | AdminUser | None:
        # if not hutils.flask.is_api_call(request.path):
        #     # if request.headers.get("Authorization"):
        #     #     return header_auth(request)
        #     # for handle new login
        #     if hutils.flask.is_login_call():
        #         # print("DDDDDDDDDDDDDDD-")
        #         return header_auth(request)

        # parse id
        acc_type, id = hutils.auth.parse_login_id(id)  # type: ignore
        if not acc_type or not id:
            return

        if acc_type == AccountType.admin:
            account = AdminUser.query.filter(AdminUser.id == id).first()
        else:
            account = User.query.filter(User.id == id).first()

        if account:
            g.account = account
            # g.account_uuid = account.uuid
            g.is_admin = hutils.flask.is_admin_role(account.role)
        return account

    @login_manager.unauthorized_handler
    def unauthorized():
        if g.user_agent['is_browser']:
            return redirect(hurl_for('common_bp.LoginView:basic_0', force=1, next={request.path}))

        else:
            abort(401, "Unauthorized")
        # return f'/{request.path.split("/")[1]}/?force=1&redirect={request.path}'

# @login_manager.request_loader



import uuid
from apiflask.fields import String, Float, Enum, Date, Integer, Boolean
from apiflask import Schema, fields
from typing import Any, Mapping

from hiddifypanel.models import UserMode, Lang, AdminMode
from hiddifypanel import hutils

# region user api


class FriendlyDateTime(fields.Field):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs):
        return hutils.convert.time_to_json(value)

    def _deserialize(self, value: Any, attr: str | None, data: Mapping[str, Any] | None, **kwargs):
        return hutils.convert.json_to_time(value)


class FriendlyUUID(fields.Field):
    def _serialize(self, value: Any, attr: str | None, obj: Any, **kwargs):
        if value is None:
            return None
        return str(value)

    def _deserialize(self, value: Any, attr: str | None, data: Mapping[str, Any] | None, **kwargs):
        if value is None:
            return None
        try:
            return uuid.UUID(value)
        except ValueError:
            self.fail('Invalid uuid')


class UserSchema(Schema):
    uuid = FriendlyUUID(required=True, description="Unique identifier for the user")
    name = String(required=True, description="Name of the user")

    usage_limit_GB = Float(
        required=False,
        allow_none=True,
        description="The data usage limit for the user in gigabytes"
    )
    package_days = Integer(
        required=False,
        allow_none=True,
        description="The number of days in the user's package"
    )
    mode = Enum(UserMode,
                required=False,
                allow_none=True,
                description="The mode of the user's account, which dictates access level or type"
                )
    last_online = FriendlyDateTime(
        format="%Y-%m-%d %H:%M:%S",
        allow_none=True,
        description="The last time the user was online, converted to a JSON-friendly format"
    )
    start_date = Date(
        format='%Y-%m-%d',
        allow_none=True,
        description="The start date of the user's package, in a JSON-friendly format"
    )
    current_usage_GB = Float(
        required=False,
        allow_none=True,
        description="The current data usage of the user in gigabytes"
    )
    last_reset_time = Date(
        format='%Y-%m-%d',
        description="The last time the user's data usage was reset, in a JSON-friendly format",
        allow_none=True
    )
    comment = String(
        missing=None,
        allow_none=True,
        description="An optional comment about the user"
    )
    added_by_uuid = FriendlyUUID(
        required=False,
        description="UUID of the admin who added this user",
        allow_none=True,
        # validate=OneOf([p.uuid for p in AdminUser.query.all()])
    )
    telegram_id = Integer(
        required=False,
        description="The Telegram ID associated with the user",
        allow_none=True
    )
    ed25519_private_key = String(
        required=False,
        allow_none=True,
        description="If empty, it will be created automatically, The user's private key using the Ed25519 algorithm"
    )
    ed25519_public_key = String(
        required=False,
        allow_none=True,
        description="If empty, it will be created automatically,The user's public key using the Ed25519 algorithm"
    )
    wg_pk = String(
        required=False,
        allow_none=True,
        description="If empty, it will be created automatically, The user's WireGuard private key"
    )

    wg_pub = String(
        required=False,
        allow_none=True,
        description="If empty, it will be created automatically, The user's WireGuard public key"
    )
    wg_psk = String(
        required=False,
        allow_none=True,
        description="If empty, it will be created automatically, The user's WireGuard preshared key"
    )

    lang = Enum(Lang, required=False, allow_none=True, description="The language of the user")


class PatchUserSchema(UserSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['name'].required = False
    pass

# endregion

# region admin api


class AdminSchema(Schema):
    name = String(required=True, description='The name of the admin')
    comment = String(required=False, description='A comment related to the admin', allow_none=True)
    uuid = FriendlyUUID(required=True, description='The unique identifier for the admin')
    mode = Enum(AdminMode, required=True, description='The mode for the admin')
    can_add_admin = Boolean(required=True, description='Whether the admin can add other admins')
    parent_admin_uuid = FriendlyUUID(description='The unique identifier for the parent admin', allow_none=True,
                                     # validate=OneOf([p.uuid for p in AdminUser.query.all()])
                                     )
    telegram_id = Integer(required=False, description='The Telegram ID associated with the admin', allow_none=True)
    lang = Enum(Lang, required=True)


class PatchAdminSchema(AdminSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['name'].required = False
        self.fields['mode'].required = False
        self.fields['lang'].required = False
        self.fields['can_add_admin'].required = False
    pass
# endregion


class SuccessfulSchema(Schema):
    status = Integer()
    msg = String()



import re
import flask_babel
import flask_babel
from flask_babel import lazy_gettext as _
# from flask_babelex import gettext as _
from flask import render_template, g  # type: ignore
from markupsafe import Markup

from hiddifypanel.hutils.flask import hurl_for
from flask import current_app as app
from hiddifypanel import hutils
from hiddifypanel.auth import login_required
import wtforms as wtf
from flask_bootstrap import SwitchField

# from gettext import gettext as _
from flask_classful import FlaskView
from flask_wtf import FlaskForm


from hiddifypanel.models import BoolConfig, StrConfig, ConfigEnum, hconfig, ConfigCategory
from hiddifypanel.models import *
from hiddifypanel.database import db
from hiddifypanel.panel import hiddify, custom_widgets
from hiddifypanel import __version__


class SettingAdmin(FlaskView):

    @login_required(roles={Role.super_admin})
    def index(self):
        form = get_config_form()
        return render_template('config.html', form=form)

    @login_required(roles={Role.super_admin})
    def post(self):
        set_hconfig(ConfigEnum.first_setup, False)
        form = get_config_form()
        reset_action = None
        if form.validate_on_submit():

            boolconfigs = BoolConfig.query.filter(BoolConfig.child_id == Child.current().id).all()
            bool_types = {c.key: 'bool' for c in boolconfigs}

            old_configs = get_hconfigs()
            changed_configs = {}

            for category, c_items in form.data.items():  # [c for c in ConfigEnum]:

                if isinstance(c_items, dict):
                    for k in ConfigEnum:
                        if k.name not in c_items:
                            continue
                        v = c_items[k.name]
                        if k.type == str:
                            if "_domain" in k or "_fakedomain" in k:
                                v = v.lower()

                            if "port" in k:
                                for p in v.split(","):
                                    for k2, v2 in c_items.items():
                                        if "port" in k2 and k.name != k2 and p in v2:
                                            hutils.flask.flash(_("Port is already used! in") + f" {k2} {k}", 'error')
                                            return render_template('config.html', form=form)
                            if k == ConfigEnum.parent_panel and v != '':
                                # v=(v+"/").replace("/admin",'')
                                v = re.sub("(/admin/.*)", "/", v)

                        if old_configs[k] != v:
                            changed_configs[k] = v

                # print(cat,vs)

            merged_configs = {**old_configs, **changed_configs}
            if len(set([merged_configs[ConfigEnum.proxy_path], merged_configs[ConfigEnum.proxy_path_client], merged_configs[ConfigEnum.proxy_path_admin]])) != 3:
                hutils.flask.flash(_("ProxyPath is already used! use different proxy path"), 'error')  # type: ignore
                return render_template('config.html', form=form)

            # validate parent_panel value
            parent_apikey = ''
            if p_p := changed_configs.get(ConfigEnum.parent_panel):
                domain, proxy_path, uuid = hutils.flask.extract_parent_info_from_url(p_p)
                if not domain or not proxy_path or not uuid or not hutils.node.is_panel_active(domain, proxy_path, uuid):
                    hutils.flask.flash(_('parent.invalid-parent-url'), 'danger')  # type: ignore
                    return render_template('config.html', form=form)
                else:
                    set_hconfig(ConfigEnum.parent_domain, domain)
                    set_hconfig(ConfigEnum.parent_admin_proxy_path, proxy_path)
                    parent_apikey = uuid

            for k, v in changed_configs.items():
                set_hconfig(k, v, commit=False)

            db.session.commit()
            flask_babel.refresh()

            # set panel mode
            p_mode = hconfig(ConfigEnum.panel_mode)
            if p_mode != PanelMode.parent:
                if hconfig(ConfigEnum.parent_panel):
                    if p_mode == PanelMode.standalone:
                        set_hconfig(ConfigEnum.panel_mode, PanelMode.child)
                else:
                    if p_mode != PanelMode.standalone:
                        set_hconfig(ConfigEnum.panel_mode, PanelMode.standalone)

            from hiddifypanel.panel.commercial.telegrambot import register_bot
            register_bot(set_hook=True)

            # sync with parent if needed
            if hutils.node.is_child():
                if hutils.node.child.is_registered():
                    if not hutils.node.child.sync_with_parent():
                        hutils.flask.flash(_('child.sync-failed'), 'danger')  # type: ignore
                    else:  # TODO: it's just for debuging
                        hutils.flask.flash(_('child.sync-success'))  # type: ignore
                else:
                    name = hconfig(ConfigEnum.unique_id)
                    parent_info = hutils.node.get_panel_info(hconfig(ConfigEnum.parent_domain), hconfig(ConfigEnum.parent_admin_proxy_path), parent_apikey)
                    if parent_info.get('version') != __version__:
                        hutils.flask.flash(_('node.diff-version'), 'danger')  # type: ignore
                    if not hutils.node.child.register_to_parent(name, parent_apikey, mode=ChildMode.remote):
                        hutils.flask.flash(_('child.register-failed'), 'danger')  # type: ignore
                    else:  # TODO: it's just for debuging
                        hutils.flask.flash(_('child.register-success'))  # type: ignore

            reset_action = hiddify.check_need_reset(old_configs)

            if old_configs[ConfigEnum.admin_lang] != hconfig(ConfigEnum.admin_lang):
                form = get_config_form()
        else:
            hutils.flask.flash(_('config.validation-error'), 'danger')  # type: ignore

        return reset_action or render_template('config.html', form=form)

    def get_babel_string(self):
        res = ""
        strconfigs = StrConfig.query.all()
        boolconfigs = BoolConfig.query.all()
        bool_types = {c.key: 'bool' for c in boolconfigs}

        configs = [*boolconfigs, *strconfigs]
        for cat in ConfigCategory:
            if cat == 'hidden':
                continue

            cat_configs = [c for c in configs if c.key.category == cat]

            for c in cat_configs:
                res += f'{{{{_("config.{c.key}.label")}}}} {{{{_("config.{c.key}.description")}}}}'

            res += f'{{{{_("config.{cat}.label")}}}}{{{{_("config.{cat}.description")}}}}'
        for c in ConfigEnum:
            res += f'{{{{_("config.{c}.label")}}}} {{{{_("config.{c}.description")}}}}'
        return res


def get_config_form():
    strconfigs = StrConfig.query.filter(StrConfig.child_id == Child.current().id).all()
    boolconfigs = BoolConfig.query.filter(BoolConfig.child_id == Child.current().id).all()
    bool_types = {c.key: 'bool' for c in boolconfigs}

    configs = [*boolconfigs, *strconfigs]
    configs_key = {k.key: k for k in configs}
    # categories=sorted([ c for c in {c.key.category:1 for c in configs}])
    # dict_configs={cat:[c for c in configs if c.category==cat] for cat in categories}

    class DynamicForm(FlaskForm):
        pass
    is_parent = hutils.node.is_parent()

    for cat in ConfigCategory:
        if cat == 'hidden':
            continue

        cat_configs = [c for c in ConfigEnum if c.category == cat and (not is_parent or c.show_in_parent)]
        if len(cat_configs) == 0:
            continue

        class CategoryForm(FlaskForm):
            description_for_fieldset = wtf.TextAreaField("", description=_(f'config.{cat}.description'), render_kw={"class": "d-none"})
        for c2 in cat_configs:
            if not (c2 in configs_key):
                continue
            c = configs_key[c2]
            if hutils.node.is_parent():
                if c.key == ConfigEnum.parent_panel:
                    continue
            extra_info = ''
            if c.key in bool_types:
                field = SwitchField(_(f'config.{c.key}.label'), default=c.value, description=_(f'config.{c.key}.description'))
            elif c.key == ConfigEnum.core_type:
                field = wtf.SelectField(_(f"config.{c.key}.label"), choices=[("xray", _("Xray")), ("singbox", _(
                    "SingBox"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.warp_mode:
                field = wtf.SelectField(
                    _(f"config.{c.key}.label"), choices=[("disable", _("Disable")), ("all", _("All")), ("custom", _("Only Blocked and Local websites"))],
                    description=_(f"config.{c.key}.description"),
                    default=hconfig(c.key))

            elif c.key == ConfigEnum.lang or c.key == ConfigEnum.admin_lang:
                field = wtf.SelectField(
                    _(f"config.{c.key}.label"),
                    choices=[("en", _("lang.en")), ("fa", Markup(_("lang.fa"))), ("zh", _("lang.zh")), ("pt", _("lang.pt")), ("ru", _("lang.ru"))],
                    description=_(f"config.{c.key}.description"),
                    default=hconfig(c.key))
            elif c.key == ConfigEnum.country:
                field = wtf.SelectField(_(f"config.{c.key}.label"), choices=[("ir", _("Iran")), ("zh", _(
                    "China")), ("other", _("Others"))], description=_(f"config.{c.key}.description"), default=hconfig(c.key))
            elif c.key == ConfigEnum.package_mode:
                package_modes = [("release", _("Release")), ("beta", _("Beta"))]
                if hconfig(c.key) == "develop":
                    package_modes.append(("develop", _("Develop")))
                field = wtf.SelectField(_(f"config.{c.key}.label"), choices=package_modes, description=_(f"config.{c.key}.description"), default=hconfig(c.key))

            # the shadowsocks2022_method is hidden now, because it only has one option to choose
            # elif c.key == ConfigEnum.shadowsocks2022_method:
            #     field = wtf.SelectField(
            #         _(f"config.{c.key}.label"),
            #         choices=[
            #             ("2022-blake3-aes-256-gcm", "2022-blake3-aes-256-gcm"),
            #             # ("2022-blake3-chacha20-poly1305", "2022-blake3-chacha20-poly1305"),
            #         ],
            #         description=_(f"config.{c.key}.description"), default=hconfig(c.key))

            elif c.key == ConfigEnum.utls:
                field = wtf.SelectField(
                    _(f"config.{c.key}.label"),
                    choices=[
                        ("none", "None"), ("chrome", "Chrome"), ("edge", "Edge"), ("ios", "iOS"), ("android", "Android"),
                        ("safari", "Safari"), ("firefox", "Firefox"), ('random', 'random'), ('randomized', 'randomized')],
                    description=_(f"config.{c.key}.description"),
                    default=hconfig(c.key)
                )
            elif c.key == ConfigEnum.telegram_lib:
                # if hconfig(ConfigEnum.telegram_lib)=='python':
                #     continue6
                libs = [
                    ("erlang", _("lib.telegram.erlang")),
                    ("python", _("lib.telegram.python")),
                    ("tgo", _("lib.telegram.go")),
                    # ("orig", _("lib.telegram.orignal")),
                ]
                field = wtf.SelectField(_("config.telegram_lib.label"), choices=libs, description=_(
                    "config.telegram_lib.description"), default=hconfig(ConfigEnum.telegram_lib))
            elif c.key == ConfigEnum.mux_protocol:
                choices = [("smux", 'smux'), ("yamux", "yamux"), ("h2mux", "h2mux")]
                field = wtf.SelectField(_(f"config.{c.key}.label"), choices=choices, description=_(f"config.{c.key}.description"), default=hconfig(c.key))

            elif c.key == ConfigEnum.warp_sites:
                validators = [wtf.validators.Length(max=2048),
                              wtf.validators.Regexp(r'^([\w.-]+)?(?:\n[\w.-]+)*$', re.IGNORECASE, _("config.invalid-pattern-for-warp-sites") + f' {c.key}')
                              ]
                render_kw = {'class': "ltr", 'maxlength': 2048}
                field = wtf.TextAreaField(_(f'config.{c.key}.label'), validators, default=c.value,
                                          description=_(f'config.{c.key}.description'), render_kw=render_kw)
            elif c.key == ConfigEnum.branding_freetext:
                validators = [wtf.validators.Length(max=2048)]
                render_kw = {'class': "ltr", 'maxlength': 2048}
                field = custom_widgets.CKTextAreaField(_(f'config.{c.key}.label'), validators, default=c.value,
                                                       description=_(f'config.{c.key}.description'), render_kw=render_kw)
            else:
                render_kw = {'class': "ltr"}
                validators = []
                if c.key == ConfigEnum.domain_fronting_domain:
                    validators.append(wtf.validators.Regexp("^([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})|$", re.IGNORECASE, _("config.Invalid domain")))
                    validators.append(hutils.flask.validate_domain_exist)
                elif '_domain' in c.key or "_fakedomain" in c.key:
                    validators.append(wtf.validators.Regexp("^([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})$", re.IGNORECASE, _("config.Invalid domain")))
                    validators.append(hutils.flask.validate_domain_exist)

                    if c.key != ConfigEnum.decoy_domain:
                        validators.append(wtf.validators.NoneOf([d.domain.lower() for d in Domain.query.all()], _("config.Domain already used")))
                        validators.append(wtf.validators.NoneOf(
                            [cc.value.lower() for cc in StrConfig.query.filter(StrConfig.child_id == Child.current().id).all() if cc.key != c.key and "fakedomain" in cc.key and cc.key != ConfigEnum.decoy_domain], _("config.Domain already used")))

                    render_kw['required'] = ""
                    if len(c.value) < 3:
                        c.value = hutils.network.get_random_domains(1)[0]

                # if c.key ==ConfigEnum.reality_short_ids:
                #     extra_info=f" <a target='_blank' href='{hurl_for('admin.Actions:get_some_random_reality_friendly_domain',test_domain=c.value)}'>"+_('Example Domains')+"</a>"
                # if c.key ==ConfigEnum.reality_server_names:
                #     validators.append(wtf.validators.Regexp("^([\w-]+\.)+[\w-]+(,\s*([\w-]+\.)+[\w-]+)*$",re.IGNORECASE,_("Invalid REALITY hostnames")))
                    # gauge width gate lamp weasel jaguar minute enough few attitude endorse situate usdt trc20 doge bep20 trx doge ltc bnb eth btc bnb
                    # enjoy control list debris chronic few door broken way negative daring life season recipe profit switch bitter casual frame aunt plate brush aerobic display

                if c.key == ConfigEnum.parent_panel:
                    validators.append(wtf.validators.Regexp("()|(http(s|)://([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})/.*)", re.IGNORECASE, _("Invalid admin link")))
                if c.key == ConfigEnum.telegram_bot_token:
                    validators.append(wtf.validators.Regexp("()|^([0-9]{8,12}:[a-zA-Z0-9_-]{30,40})|$", re.IGNORECASE, _("config.Invalid telegram bot token")))
                if c.key == ConfigEnum.branding_site:
                    validators.append(wtf.validators.Regexp(
                        "()|(http(s|)://([A-Za-z0-9\\-\\.]+\\.[a-zA-Z]{2,})/?.*)", re.IGNORECASE, _("config.Invalid brand link")))
                    # render_kw['required']=""

                if 'secret' in c.key:
                    validators.append(wtf.validators.Regexp(
                        "^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", re.IGNORECASE, _('config.invalid uuid')))
                    render_kw['required'] = ""

                if c.key == ConfigEnum.proxy_path:
                    validators.append(wtf.validators.Regexp("^[a-zA-Z0-9]*$", re.IGNORECASE, _("config.Invalid proxy path")))
                    render_kw['required'] = ""

                if 'port' in c.key:
                    if c.key in [ConfigEnum.http_ports, ConfigEnum.tls_ports]:
                        validators.append(wtf.validators.Regexp("^(\\d+)(,\\d+)*$", re.IGNORECASE, _("config.Invalid port")))
                        render_kw['required'] = ""
                    else:
                        validators.append(wtf.validators.Regexp("^(\\d+)(,\\d+)*$|^$", re.IGNORECASE, _("config.Invalid port")))
                    # validators.append(wtf.validators.Regexp("^(\d+)(,\d+)*$",re.IGNORECASE,_("config.port is required")))

                # tls tricks validations
                if c.key in [ConfigEnum.tls_fragment_size, ConfigEnum.tls_fragment_sleep, ConfigEnum.tls_padding_length, ConfigEnum.wireguard_noise_trick]:
                    validators.append(wtf.validators.Regexp("^\\d+-\\d+$", re.IGNORECASE, _("config.Invalid! The pattern is number-number") + f' {c.key}'))
                # mux and hysteria validations
                if c.key in [ConfigEnum.hysteria_up_mbps, ConfigEnum.hysteria_down_mbps, ConfigEnum.mux_max_connections, ConfigEnum.mux_min_streams, ConfigEnum.mux_max_streams,
                             ConfigEnum.mux_brutal_down_mbps, ConfigEnum.mux_brutal_up_mbps]:
                    validators.append(wtf.validators.Regexp("^\\d+$", re.IGNORECASE, _("config.Invalid! it should be a number only") + f' {c.key}'))

                for val in validators:
                    if hasattr(val, "regex"):
                        render_kw['pattern'] = val.regex.pattern
                        render_kw['title'] = val.message

                if c.key == ConfigEnum.reality_public_key and g.account.mode in [AdminMode.super_admin]:
                    extra_info = f" <a href='{hurl_for('admin.Actions:change_reality_keys')}'>{_('Change')}</a>"

                field = wtf.StringField(_(f'config.{c.key}.label'), validators, default=c.value,
                                        description=_(f'config.{c.key}.description') + extra_info, render_kw=render_kw)
            setattr(CategoryForm, f'{c.key}', field)

        multifield = wtf.FormField(CategoryForm, Markup('<i class="fa-solid fa-plus"></i>&nbsp' + _(f'config.{cat}.label')))

        setattr(DynamicForm, cat, multifield)

    setattr(DynamicForm, "submit", wtf.SubmitField(_('Submit')))

    return DynamicForm()



from enum import auto
from typing import List

from urllib.parse import urlparse

from flask import request

from flask_babel import lazy_gettext as _
from sqlalchemy.orm import backref
from urllib.parse import urlparse
from strenum import StrEnum

from hiddifypanel.database import db

from hiddifypanel.models.config import hconfig
from .child import Child
from hiddifypanel.models.config_enum import ConfigEnum
from sqlalchemy.orm import backref
from sqlalchemy_serializer import SerializerMixin


class DomainType(StrEnum):
    direct = auto()
    sub_link_only = auto()
    cdn = auto()
    auto_cdn_ip = auto()
    relay = auto()
    reality = auto()
    old_xtls_direct = auto()
    worker = auto()
    fake = auto()

    # fake_cdn = "fake_cdn"
    # telegram_faketls = "telegram_faketls"
    # ss_faketls = "ss_faketls"


ShowDomain = db.Table('show_domain',
                      db.Column('domain_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True),
                      db.Column('related_id', db.Integer, db.ForeignKey('domain.id'), primary_key=True)
                      )


class Domain(db.Model, SerializerMixin):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    child_id = db.Column(db.Integer, db.ForeignKey('child.id'), default=0)
    domain = db.Column(db.String(200), nullable=False, unique=False)
    alias = db.Column(db.String(200))
    sub_link_only = db.Column(db.Boolean, nullable=False, default=False)
    mode = db.Column(db.Enum(DomainType), nullable=False, default=DomainType.direct)
    cdn_ip = db.Column(db.Text(2000), nullable=True, default='')
    # port_index=db.Column(db.Integer, nullable=True, default=0)
    grpc = db.Column(db.Boolean, nullable=True, default=False)
    servernames = db.Column(db.String(1000), nullable=True, default='')
    # show_all=db.Column(db.Boolean, nullable=True)
    show_domains = db.relationship('Domain', secondary=ShowDomain,
                                   primaryjoin=id == ShowDomain.c.domain_id,
                                   secondaryjoin=id == ShowDomain.c.related_id,
                                   backref=backref('showed_by_domains', lazy='dynamic')
                                   )
    extra_params = db.Column(db.String(200), nullable=True, default='')

    def __repr__(self):
        return f'{self.domain}'

    def to_dict(self, dump_ports=False, dump_child_id=False):
        data = {
            'domain': self.domain.lower(),
            'mode': self.mode,
            'alias': self.alias,
            'sub_link_only': self.sub_link_only,
            'child_unique_id': self.child.unique_id if self.child else '',
            'cdn_ip': self.cdn_ip,
            'servernames': self.servernames,
            'grpc': self.grpc,
            'show_domains': [dd.domain for dd in self.show_domains],
        }
        if dump_child_id:
            data['child_id'] = self.child_id
        if dump_ports:
            data["internal_port_hysteria2"] = self.internal_port_hysteria2
            data["internal_port_tuic"] = self.internal_port_tuic
            data["internal_port_reality"] = self.internal_port_reality
            data["need_valid_ssl"] = self.need_valid_ssl

        return data

    @staticmethod
    def from_schema(schema):
        return schema.dump(Domain())

    def to_schema(self):
        domain_dict = self.to_dict()
        from hiddifypanel.panel.commercial.restapi.v2.parent.schema import DomainSchema
        return DomainSchema().load(domain_dict)

    @property
    def need_valid_ssl(self):
        return self.mode in [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]

    @property
    def port_index(self):
        return self.id

    @property
    def internal_port_hysteria2(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        # print("child_id",self.child_id)
        return int(hconfig(ConfigEnum.hysteria_port, self.child_id)) + self.port_index

    @property
    def internal_port_tuic(self):
        if self.mode not in [DomainType.direct, DomainType.relay, DomainType.fake]:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.tuic_port, self.child_id)) + self.port_index

    @property
    def internal_port_reality(self):
        if self.mode != DomainType.reality:
            return 0
        # TODO: check validity of the range of the port
        return int(hconfig(ConfigEnum.reality_port, self.child_id)) + self.port_index


def hdomains(mode):
    domains = Domain.query.filter(Domain.mode == mode).all()
    if domains:
        return [d.domain for d in domains]
    return []


def hdomain(mode):
    domains = hdomains(mode)
    if domains:
        return domains[0]
    return None


def get_hdomains():
    return {mode: hdomains(mode) for mode in DomainType}


def get_domain(domain):
    return Domain.query.filter(Domain.domain == domain).first()


def get_panel_link(child_id: int | None = None) -> Domain | None:
    if child_id is None:
        child_id = Child.current().id
    domains = Domain.query.filter(Domain.mode.in_(
        [DomainType.direct, DomainType.cdn, DomainType.worker, DomainType.relay, DomainType.auto_cdn_ip, DomainType.old_xtls_direct, DomainType.sub_link_only]),
        Domain.child_id == child_id
    ).all()
    if not domains:
        return None
    return domains[0]


def get_panel_domains(always_add_ip=False, always_add_all_domains=False) -> List[Domain]:
    from hiddifypanel import hutils
    domains = []
    domains = Domain.query.filter(Domain.mode == DomainType.sub_link_only, Domain.child_id == Child.current().id).all()
    if not len(domains) or always_add_all_domains:
        domains = Domain.query.filter(Domain.mode.notin_([DomainType.fake, DomainType.reality])).all()

    if len(domains) == 0 and request:
        domains = [Domain(domain=request.host)]
    if len(domains) == 0 or always_add_ip:
        domains += [Domain(domain=hutils.network.get_ip_str(4))]
    return domains


def get_proxy_domains(domain):
    db_domain = Domain.query.filter(Domain.domain == domain, Domain.child_id == Child.current().id).first()
    if not db_domain:
        db_domain = Domain(domain=domain, mode=DomainType.direct, cdn_ip='', show_domains=[])
    return get_proxy_domains_db(db_domain)


def get_proxy_domains_db(db_domain):

    if not db_domain:
        domain = urlparse(request.base_url).hostname
        db_domain = Domain(domain=domain, mode=DomainType.direct, show_domains=[])
        # print("no domain")
        hutils.flask.flash(_("This domain does not exist in the panel!" + domain))  # type: ignore

    return db_domain.show_domains or Domain.query.all()


def get_current_proxy_domains(force_domain=None):
    domain = force_domain or urlparse(request.base_url).hostname
    return get_proxy_domains(domain)


def add_or_update_domain(commit=True, child_id=0, **domain):
    dbdomain = Domain.query.filter(Domain.domain == domain['domain']).first()
    if not dbdomain:
        dbdomain = Domain(domain=domain['domain'])
        db.session.add(dbdomain)
    dbdomain.child_id = child_id

    dbdomain.mode = domain['mode']
    if (str(domain.get('sub_link_only', False)).lower() == 'true'):
        dbdomain.mode = DomainType.sub_link_only
    dbdomain.cdn_ip = domain.get('cdn_ip', '')
    dbdomain.alias = domain.get('alias', '')
    dbdomain.grpc = domain.get('grpc', False)
    dbdomain.servernames = domain.get('servernames', '')
    show_domains = domain.get('show_domains', [])
    dbdomain.show_domains = Domain.query.filter(Domain.domain.in_(show_domains)).all()
    if commit:
        db.session.commit()


def bulk_register_domains(domains, commit=True, remove=False, force_child_unique_id: str | None = None):
    from hiddifypanel.panel import hiddify
    child_ids = {}
    for domain in domains:
        child_id = hiddify.get_child(unique_id=force_child_unique_id)
        child_ids[child_id] = 1
        add_or_update_domain(commit=False, child_id=child_id, **domain)
    if remove and len(child_ids):
        dd = {d['domain']: 1 for d in domains}
        for d in Domain.query.filter(Domain.child_id.in_(child_ids)):
            if d.domain not in dd:
                db.session.delete(d)

    if commit:
        db.session.commit()



from apiflask import APIBlueprint, Schema
from apiflask.fields import Integer, String
from flask import g
from hiddifypanel.models import AdminUser

bp = APIBlueprint("api_admin", __name__, url_prefix="/<proxy_path>/api/v2/admin/", enable_openapi=True)


def init_app(app):

    with app.app_context():
        from .admin_info_api import AdminInfoApi
        from .server_status_api import AdminServerStatusApi
        from .admin_user_api import AdminUserApi
        from .admin_users_api import AdminUsersApi
        from .admin_log_api import AdminLogApi
        bp.add_url_rule('/me/', view_func=AdminInfoApi)
        bp.add_url_rule('/server_status/', view_func=AdminServerStatusApi)
        bp.add_url_rule('/admin_user/<uuid:uuid>/', view_func=AdminUserApi)
        bp.add_url_rule('/admin_user/', view_func=AdminUsersApi)
        bp.add_url_rule('/log/', view_func=AdminLogApi)

        from .user_api import UserApi
        from .users_api import UsersApi
        bp.add_url_rule('/user/<uuid:uuid>/', view_func=UserApi)
        bp.add_url_rule('/user/', view_func=UsersApi)
    app.register_blueprint(bp)


def has_permission(model) -> bool:
    '''Check if the authenticated account has permission to do an action(get,insert,update,delete) on the another admin'''
    if not g.account.uuid != AdminUser.get_super_admin_uuid() and model.added_by != g.account.id:  # type: ignore
        return False
    return True



from flask import current_app as app
from flask import g
from apiflask import abort
from flask.views import MethodView
from hiddifypanel.auth import login_required
from hiddifypanel.models.role import Role
from .admin_user_api import AdminSchema, has_permission
from hiddifypanel.models import AdminUser


class AdminUsersApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin})]

    @app.output(AdminSchema(many=True))  # type: ignore
    def get(self):
        admins = AdminUser.query.filter(AdminUser.id.in_(g.account.recursive_sub_admins_ids())).all() or abort(404, "You have no admin")
        return [admin.to_dict() for admin in admins]  # type: ignore

    @app.input(AdminSchema, arg_name='data')  # type: ignore
    @app.output(AdminSchema)  # type: ignore
    def put(self, data):
        uuid = data.get('uuid') or abort(422, "Parameter issue: 'uuid'")
        admin = AdminUser.by_uuid(uuid)  # type: ignore

        if not data.get('added_by_uuid'):
            data['added_by_uuid'] = g.account.uuid

        # update check permission
        if admin:
            if not has_permission(admin):
                abort(403, "You don't have permission to access this admin")

        dbadmin = AdminUser.add_or_update(**data) or abort(502, "Unknown issue: Admin is not added")
        return dbadmin.to_dict()  # type: ignore



from .VERSION import __version__, __release_date__
# from . import cache
from . import Events
from .base import create_app, create_app_wsgi
# from . import hutils
# from . import panel


__all__ = ["create_app", "create_app_wsgi"]

# application = create_app_wsgi()



from apiflask import Schema, fields


class RegisterWithParentInputSchema(Schema):
    parent_panel = fields.String(required=True, description="The parent panel url")
    name = fields.String(required=True, description="The child's name in the parent panel")
    apikey = fields.String(description="The parent's apikey")



from flask import g
from flask.views import MethodView
from flask import current_app as app
from apiflask import abort
from hiddifypanel.auth import login_required
from hiddifypanel.models import *
from hiddifypanel.panel import hiddify
from hiddifypanel.drivers import user_driver
from hiddifypanel.panel import hiddify

from . import has_permission
from .schema import UserSchema, PatchUserSchema, SuccessfulSchema


class UserApi(MethodView):
    decorators = [login_required({Role.super_admin, Role.admin, Role.agent})]

    @app.output(UserSchema)  # type: ignore
    def get(self, uuid):
        user = User.by_uuid(uuid) or abort(404, "user not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")

        return user.to_dict(False)  # type: ignore

    @app.input(PatchUserSchema, arg_name="data")  # type: ignore
    @app.output(SuccessfulSchema)  # type: ignore
    def patch(self, uuid, data):
        user = User.by_uuid(uuid) or abort(404, "user not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")

        data['uuid'] = uuid
        if not data.get('added_by_uuid'):
            data['added_by_uuid'] = g.account.uuid

        User.add_or_update(**data)  # type: ignore
        user = User.by_uuid(uuid) or abort(502, "unknown issue! user is not added")
        user_driver.add_client(user)
        hiddify.quick_apply_users()
        return {'status': 200, 'msg': 'ok'}

    @app.output(SuccessfulSchema)  # type: ignore
    def delete(self, uuid):
        user = User.by_uuid(uuid) or abort(404, "user not found")
        if not has_permission(user):
            abort(403, "You don't have permission to access this user")
        user.remove()  # type: ignore
        hiddify.quick_apply_users()
        return {'status': 200, 'msg': 'ok'}




'''This file isn't a model, it's just an enum for roles'''
from strenum import StrEnum


class Role(StrEnum):
    super_admin = "super_admin"
    admin = "admin"
    agent = "agent"
    user = "user"


class AccountType(StrEnum):
    admin = "admin"
    user = "user"



import telebot
from flask import request
from apiflask import abort
from flask_restful import Resource

from hiddifypanel.models import *
from hiddifypanel import Events
logger = telebot.logger


class ExceptionHandler(telebot.ExceptionHandler):
    def handle(self, exception):
        logger.error(exception)


bot = telebot.TeleBot("", parse_mode="HTML", threaded=False, exception_handler=ExceptionHandler())
bot.username = ''


def register_bot(set_hook=False):
    try:
        global bot
        token = hconfig(ConfigEnum.telegram_bot_token)
        if token:
            bot.token = hconfig(ConfigEnum.telegram_bot_token)
            try:
                bot.username = bot.get_me().username
            except BaseException:
                pass
            # bot.remove_webhook()
            # time.sleep(0.1)
            domain = get_panel_domains()[0].domain
            admin_proxy_path = hconfig(ConfigEnum.proxy_path_admin)

            user_secret = AdminUser.get_super_admin_uuid()
            if set_hook:
                bot.set_webhook(url=f"https://{domain}/{admin_proxy_path}/{user_secret}/api/v1/tgbot/")
    except Exception as e:
        print(e)
        import traceback
        traceback.print_stack()


def init_app(app):
    with app.app_context():
        global bot
        token = hconfig(ConfigEnum.telegram_bot_token)
        if token:
            bot.token = token
            try:
                bot.username = bot.get_me().username
            except BaseException:
                pass


class TGBotResource(Resource):
    def post(self):
        try:
            if request.headers.get('content-type') == 'application/json':
                json_string = request.get_data().decode('utf-8')
                update = telebot.types.Update.de_json(json_string)
                bot.process_new_updates([update])
                return ''
            else:
                abort(403)
        except Exception as e:
            print("Error", e)
            import traceback
            traceback.print_exc()
            return "", 500
